import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pdb

#possUtil/HPUS/simulation(possession and full match)/momentum indicator/heatmap

def cal_poss_util(data, shot_num=[6,8], cross_num=[4], num_actions=9):
    # Check if the input is a DataFrame
    if not isinstance(data, pd.DataFrame):
        data = pd.read_csv(data)
    
    # Map actions to integers
    action_dict = {'short_pass': 0, 'carry': 1, 'high_pass': 2, '_': 3, 'cross': 4, 
                   'long_pass': 5, 'shot': 6, 'dribble': 7}
    data["action"] = data["action"].map(action_dict)
    data.loc[(data["action"] == 6) & (data["goal"] == 1), "action"] = 8

    #drop all row with action = period_over
    data = data[data["action"] != 3]

    # Prepare action probability columns
    action_columns = [f'action_{i}_prob' for i in range(num_actions)]
    
    # Ensure action probabilities sum to 1 (apply softmax)
    row_sums = data[action_columns].sum(axis=1)
    if not np.allclose(row_sums, 1):
        exp_values = np.exp(data[action_columns])
        data[action_columns] = exp_values.div(exp_values.sum(axis=1), axis=0)

    # Fill NaN values with 0
    data.fillna(0, inplace=True)

    # Sum up shot and cross probabilities
    poss_util_prob = data[[f'action_{i}_prob' for i in shot_num + cross_num]].sum(axis=1)
    data['poss_util_prob'] = poss_util_prob

    # Compute attack_flag efficiently
    relevant_actions = shot_num + cross_num
    data['attack_flag'] = data['action'].isin(relevant_actions).astype(int)
    
    # Group by match_id and poss_id, and calculate total probabilities
    poss_util = data.groupby(['match_id', 'poss_id'], as_index=False).agg({
        'poss_util_prob': 'sum',
        'attack_flag': 'sum',
        'team': 'first',
    })
    
    # Set poss_util_prob to negative if no shot or cross occurred
    poss_util.loc[poss_util['attack_flag'] == 0, 'poss_util_prob'] *= -1

    # Drop attack_flag from the final output
    poss_util.drop(columns='attack_flag', inplace=True)
    
    return poss_util

def plot_poss_util_dist(poss_util, save_path, bins=20):
    # Get the unique team IDs
    teams = poss_util['team'].unique()
    
    # Set up the figure
    plt.figure(figsize=(12, 8))

    # Plot a density curve for each team
    for team in teams:
        team_data = poss_util[poss_util['team'] == team]['poss_util_prob']
        
        # Calculate the kernel density estimation using a histogram
        density, bins = np.histogram(team_data, bins=bins, density=True)
        bin_centers = 0.5 * (bins[1:] + bins[:-1])
        
        # Plot the density curve
        plt.plot(bin_centers, density, label=team, linewidth=2)

    # Set plot labels and title with increased font size
    plt.xlabel('Poss-Util Probability', fontsize=18)  # Increased to 16
    plt.ylabel('Density', fontsize=18)  # Increased to 16
    plt.title('Poss-Util', fontsize=18)  # Increased to 18
    
    # Increase font size for tick labels
    plt.tick_params(axis='both', labelsize=16)  # Set font size for x and y tick labels

    # Add legend with increased font size
    plt.legend(title='Team ID', fontsize=16)  # Increased to 14
    
    # Display the plot and save it
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.tight_layout()
    plt.savefig(save_path + "poss_util.png")
    plt.close()  # Close the plot to free memory

def plot_poss_util_plus_dist(poss_util, save_path, bins = 20):
    #for each team summarize the poss_util_prob (only positive value) per match_id
    # Get the unique team IDs
    teams = poss_util['team'].unique()
    poss_util_plus = poss_util[poss_util['poss_util_prob'] > 0]
    poss_util_plus = poss_util_plus.groupby(['team','match_id'],as_index=False).agg({'poss_util_prob':'sum'})
    # Set up the figure
    plt.figure(figsize=(12, 8))
    for team in teams:
        team_data = poss_util_plus[poss_util_plus['team'] == team]['poss_util_prob']
        
        # Calculate the kernel density estimation using a histogram
        density, bins = np.histogram(team_data, bins=bins, density=True)
        bin_centers = 0.5 * (bins[1:] + bins[:-1])
        
        # Plot the density curve
        plt.plot(bin_centers, density, label=team, linewidth=2)
        # Plot the mean value
        plt.axvline(x=team_data.mean(), color='black', linestyle='--', linewidth=2)

    # Set plot labels and title with increased font size
    plt.xlabel('Poss-Util+ per Match', fontsize=18)  # Increased to 16
    plt.ylabel('Density', fontsize=18)  # Increased to 16
    plt.title('Poss-Util+', fontsize=18)  # Increased to 18
    # Increase font size for tick labels
    plt.tick_params(axis='both', labelsize=16)  # Set font size for x and y tick labels
    # Add legend with increased font size
    plt.legend(title='Team ID', fontsize=16)  # Increased to 14
    # Display the plot and save it
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.tight_layout()
    plt.savefig(save_path + "poss_util_plus.png")
    plt.close()  # Close the plot to free memory

if __name__ == "__main__":
    import os
    inference_data = os.getcwd()+"/test/inference/nmstpp/inference.csv"
    save_path = os.getcwd()+"/test/application/"
    poss_util = cal_poss_util(inference_data)
    poss_util.to_csv(save_path+"poss_util.csv",index=False)
    plot_poss_util_dist(poss_util,save_path,bins=20)
    plot_poss_util_plus_dist(poss_util,save_path,bins=20)
    






import torch
from ..utils.UEID_preprocessing import UEID_preprocessing
from ..dataloaders.UEID_loader import UEID_loader
from ..dataloaders.UEID_loader import UEID_Simulation_loader
from ..utils.LEM_cost_function import LEM_cost_function
from ..trainers.LEM_train import one_hot_encode_from_continuous
from torch.utils.data import DataLoader
import json
import pandas as pd
import numpy as np
from tqdm import tqdm
import os
import pdb


def LEM_inference(train_df, data, model_name, model_path, model_config, num_workers=4, batch_size=64, device=None, min_max_dict_path=None):
    # Load the training configuration 
    with open(model_config, "r") as f:
        config = json.load(f)
    
    #check if train_df and data are paths or dataframes
    if train_df is not None and train_df != "None":
        if not isinstance(train_df, pd.DataFrame):
            #check if the train_df is a path
            if isinstance(train_df, str):
                train_df = pd.read_csv(train_df) if not config['test'] else pd.read_csv(train_df).head(1000)
            else:
                raise ValueError("train_df must be a pandas dataframe or a path to a csv file")
    if isinstance(data, pd.DataFrame):
        data_load = data
    elif isinstance(data, str):
        data_load = pd.read_csv(data) if not config['test'] else pd.read_csv(data).head(1000)
    else:
        raise ValueError("data must be a pandas dataframe or a path to a csv file")

    data = data_load.copy()

    #set the hyperparameters
    seq_len = config["seq_len"]
    features = config["features"]
    num_workers = config["dataloader_num_worker"]
    batch_size = config["batch_size"]
    device = config["device"]
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device
    
    # Preprocess the data
    if min_max_dict_path != "None" and min_max_dict_path is not None:
        try:
            with open(min_max_dict_path, "r") as f:
                min_max_dict = json.load(f)
            min_dict = min_max_dict['min_dict']
            max_dict = min_max_dict['max_dict']
        except:
            train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    else:
        train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    data, _, _ = UEID_preprocessing(data, min_dict, max_dict)

    # Create the data loader
    data_loader = UEID_loader(data, seq_len, features)
    data_loader = DataLoader(data_loader, batch_size=batch_size, num_workers=num_workers, shuffle=False, drop_last=False)

    # Load the model
    model = load_model(model_name, model_path, model_config)
    model.load_state_dict(torch.load(model_path, weights_only=True))
    model.to(device)
    model.eval()
    
    # Run inference
    total_loss = 0
    total_loss_components = [0]*5
    with torch.no_grad():
        predicted = []
        end_idx_list = []
        for i, (input_seq, gt, end_idx) in tqdm(enumerate(data_loader)):
            input_seq, gt = input_seq.to(device), gt.to(device)
            num_action=config["num_actions"] if config is not None and config !="None" else 9
            inputs_one_hot = torch.nn.functional.one_hot(input_seq[:,:,0].long(), num_classes=num_action).float()
            input_seq = torch.cat((inputs_one_hot, input_seq[:,:,1:]), dim=2)
            batch_size = input_seq.shape[0]
            input_seq = input_seq.reshape(batch_size, -1)

            gt_action_one_hot = torch.nn.functional.one_hot(gt[:,0].long(), num_classes=num_action).float()
            gt_deltaT_one_hot = one_hot_encode_from_continuous(gt[:,1], num_bins=config["delta_T_bin"]).float()
            gt_start_x_one_hot = one_hot_encode_from_continuous(gt[:,2], num_bins=config["start_x_bin"]).float()
            gt_start_y_one_hot = one_hot_encode_from_continuous(gt[:,3], num_bins=config["start_y_bin"]).float()
            gt = torch.cat((gt_action_one_hot, gt_deltaT_one_hot, gt_start_x_one_hot, gt_start_y_one_hot), dim=1)
            output = model(input_seq)
            #check if output is in device
            if output.device != device:
                output = output.to(device)
            loss, loss_components= LEM_cost_function(gt, output,  min_dict=min_dict, max_dict=max_dict, device=device,config=config)
            total_loss += loss.item()
            loss_components = [i.cpu().detach().numpy() for i in loss_components] if device != "cpu" else [i.detach().numpy() for i in loss_components]
            total_loss_components = [total_loss_components[j] + loss_components[j] for j in range(len(loss_components))]
            if device != "cpu":
                predicted.append(output.cpu().numpy())
                end_idx_list.append(end_idx.cpu().numpy())
            else:
                predicted.append(output.numpy())
                end_idx_list.append(end_idx.numpy())
        # Concatenate the results
        predicted = np.concatenate(predicted, axis=0)
        end_idx_list = np.concatenate(end_idx_list, axis=0)
        average_loss = total_loss / len(data_loader)
        average_loss_components = [total_loss_components[j] / len(data_loader) for j in range(len(loss_components))]

    #save the loss_df
    columns= ['train_loss', 'BCEL_continuous', 'MAE_deltaT', 'MAE_start_x', 'MAE_start_y']
    loss_df = pd.DataFrame([[average_loss]+average_loss_components], columns=columns)
    loss_df = loss_df.round(4)
        
    data_raw = data_load.copy()

    new_data = {
        'action_pred': [],
        'delta_T_pred': [],
        'start_x_pred': [],
        'start_y_pred': [],
        'delta_T_pred_unscaled': [],
        'start_x_pred_unscaled': [],
        'start_y_pred_unscaled': []
    }

    # Add probability columns
    for action_i in range(config['num_actions']):
        new_data[f'action_{action_i}_prob'] = []
    for delta_T_i in range(config['delta_T_bin']):
        new_data[f'delta_T_{delta_T_i}_prob'] = []
    for start_x_i in range(config['start_x_bin']):
        new_data[f'start_x_{start_x_i}_prob'] = []
    for start_y_i in range(config['start_y_bin']):
        new_data[f'start_y_{start_y_i}_prob'] = []

    # Process predictions
    for i, idx in enumerate(end_idx_list):
        new_data['action_pred'].append(np.argmax(predicted[i, :config['num_actions']]))
        new_data['delta_T_pred'].append(np.argmax(predicted[i, config['num_actions']:config['num_actions']+config['delta_T_bin']])/100)
        new_data['start_x_pred'].append(np.argmax(predicted[i, config['num_actions']+config['delta_T_bin']:config['num_actions']+config['delta_T_bin']+config['start_x_bin']])/100)
        new_data['start_y_pred'].append(np.argmax(predicted[i, config['num_actions']+config['delta_T_bin']+config['start_x_bin']:])/100)

        # Probability columns
        for action_i in range(config['num_actions']):
            new_data[f'action_{action_i}_prob'].append(predicted[i, action_i])
        for delta_T_i in range(config['delta_T_bin']):
            new_data[f'delta_T_{delta_T_i}_prob'].append(predicted[i, delta_T_i+config['num_actions']])
        for start_x_i in range(config['start_x_bin']):
            new_data[f'start_x_{start_x_i}_prob'].append(predicted[i, start_x_i+config['num_actions']+config['delta_T_bin']])
        for start_y_i in range(config['start_y_bin']):
            new_data[f'start_y_{start_y_i}_prob'].append(predicted[i, start_y_i+config['num_actions']+config['delta_T_bin']+config['start_x_bin']])

        # Unscaled predictions
        new_data['delta_T_pred_unscaled'].append(np.exp(np.argmax(predicted[i, config['num_actions']:config['num_actions']+config['delta_T_bin']])/100*(max_dict["delta_T"]-min_dict["delta_T"]) + min_dict["delta_T"]) - 1e-6)
        new_data['start_x_pred_unscaled'].append(np.argmax(predicted[i, config['num_actions']+config['delta_T_bin']:config['num_actions']+config['delta_T_bin']+config['start_x_bin']])/100*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"])
        new_data['start_y_pred_unscaled'].append(np.argmax(predicted[i, config['num_actions']+config['delta_T_bin']+config['start_x_bin']:])/100*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"])

    # Create a new DataFrame with all the new data
    new_df = pd.DataFrame(new_data, index=end_idx_list)
    df = data_raw.join(new_df)

    return df, loss_df

def LEM_simulation_possession(train_df, data, model_name, model_path, model_config, num_workers=4, batch_size=64, device=None, random_selection=False, max_iter=200, min_max_dict_path=None):
        # Load the training configuration 
    with open(model_config, "r") as f:
        config = json.load(f)
    
    #check if train_df and data are paths or dataframes
    if train_df is not None and train_df != "None":
        if not isinstance(train_df, pd.DataFrame):
            #check if the train_df is a path
            if isinstance(train_df, str):
                train_df = pd.read_csv(train_df) if not config['test'] else pd.read_csv(train_df).head(1000)
            else:
                raise ValueError("train_df must be a pandas dataframe or a path to a csv file")
    if isinstance(data, pd.DataFrame):
        data_raw = data
    elif isinstance(data, str):
        data_raw = pd.read_csv(data) if not config['test'] else pd.read_csv(data).head(1000)
    else:
        raise ValueError("data must be a pandas dataframe or a path to a csv file")
    data = data_raw.copy()

    #set the hyperparameters
    seq_len = config["seq_len"]
    features = config["features"]
    num_workers = config["dataloader_num_worker"]
    batch_size = config["batch_size"]
    device = config["device"]
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device
    
    # Preprocess the data
    if min_max_dict_path != "None" and min_max_dict_path is not None:
        try:
            with open(min_max_dict_path, "r") as f:
                min_max_dict = json.load(f)
            min_dict = min_max_dict['min_dict']
            max_dict = min_max_dict['max_dict']
        except:
            train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    else:
        train_df, min_dict, max_dict = UEID_preprocessing(train_df)

    data, _, _ = UEID_preprocessing(data, min_dict, max_dict)

    # Create the data loader
    data_loader = UEID_Simulation_loader(data, seq_len, features)
    data_loader = DataLoader(data_loader, batch_size=batch_size, num_workers=num_workers, shuffle=False, drop_last=False)

    # Load the model
    model = load_model(model_name, model_path, model_config)
    model.load_state_dict(torch.load(model_path, weights_only=True))
    model.to(device)
    model.eval()

    # Run inference
    with torch.no_grad():
        simulation = {}
        for i, (input_seq, end_idx) in tqdm(enumerate(data_loader)):
            #convert the end_idx to a list
            end_idx = end_idx.cpu().numpy().tolist() if device != "cpu" else end_idx.numpy().tolist()

            for idx in end_idx:
                simulation[idx] = []
            num_action=config["num_actions"] if config is not None and config !="None" else 9
            inputs_one_hot = torch.nn.functional.one_hot(input_seq[:,:,0].long(), num_classes=num_action).float()
            input_seq = torch.cat((inputs_one_hot, input_seq[:,:,1:]), dim=2)
            batch_size = input_seq.shape[0]
            input_seq = input_seq.reshape(batch_size, -1)
            input_seq = input_seq.to(device)
            iteration = 1

            # Run the simulation
            while end_idx:

                # Forward pass through the model
                output = model(input_seq)

                # Convert the first config['num_actions'] to action probabilities
                action_probs = torch.nn.functional.softmax(output[:, :config['num_actions']], dim=1)
                delta_T_probs = torch.nn.functional.softmax(output[:, config['num_actions']:config['num_actions']+config['delta_T_bin']], dim=1)
                start_x_probs = torch.nn.functional.softmax(output[:, config['num_actions']+config['delta_T_bin']:config['num_actions']+config['delta_T_bin']+config['start_x_bin']], dim=1)
                start_y_probs = torch.nn.functional.softmax(output[:, config['num_actions']+config['delta_T_bin']+config['start_x_bin']:], dim=1)

                # Select actions based on the probabilities
                if random_selection:
                    action = torch.multinomial(action_probs, 1)
                    delta_T = torch.multinomial(delta_T_probs, 1)
                    start_x = torch.multinomial(start_x_probs, 1)
                    start_y = torch.multinomial(start_y_probs, 1)
                    action_prob = torch.gather(action_probs, 1, action)
                    delta_T_prob = torch.gather(delta_T_probs, 1, delta_T)
                    start_x_prob = torch.gather(start_x_probs, 1, start_x)
                    start_y_prob = torch.gather(start_y_probs, 1, start_y)
                    delta_T = delta_T/100
                    start_x = start_x/100
                    start_y = start_y/100
                else:
                    action = torch.argmax(action_probs, dim=1)
                    action = action.unsqueeze(1)
                    action_prob = torch.gather(action_probs, 1, action)
                    delta_T = torch.argmax(delta_T_probs, dim=1)
                    delta_T = delta_T.unsqueeze(1)
                    delta_T_prob = torch.gather(delta_T_probs, 1, delta_T)
                    delta_T = delta_T/100
                    start_x = torch.argmax(start_x_probs, dim=1)
                    start_x = start_x.unsqueeze(1)
                    start_x_prob = torch.gather(start_x_probs, 1, start_x)
                    start_x = start_x/100
                    start_y = torch.argmax(start_y_probs, dim=1)
                    start_y = start_y.unsqueeze(1)
                    start_y_prob = torch.gather(start_y_probs, 1, start_y)
                    start_y = start_y/100
                    
                # Recombine the action with the delta_T, start_x, start_y
                
                output = torch.cat((action.float(), delta_T.float(), start_x.float(), start_y.float(), action_prob.float(), delta_T_prob.float(), start_x_prob.float(), start_y_prob.float()), dim=1)

                # Update the simulation dictionary
                mask = torch.ones(len(end_idx), dtype=bool)  # Initialize a mask to keep all entries

                for j, idx in enumerate(end_idx):
                    output_j = output[j].cpu().numpy().tolist() if device != "cpu" else output[j].numpy().tolist()
                    action_j = int(output_j[0])
                    delta_T_j = np.exp(output_j[1]*(max_dict["delta_T"]-min_dict["delta_T"]) + min_dict["delta_T"]) - 1e-6
                    start_x_j = output_j[2]*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"]
                    start_y_j = output_j[3]*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"]
                    
                    action_prob_j = output_j[4]
                    delta_T_prob_j = output_j[5]
                    start_x_prob_j = output_j[6]
                    start_y_prob_j = output_j[7]

                    simulation[idx].append([iteration, action_j, action_prob_j, delta_T_j, delta_T_prob_j, start_x_j, start_x_prob_j, start_y_j, start_y_prob_j])

                    # Check if the current action indicates termination
                    if output[j, 0] == 3:
                        mask[j] = False  # Mark as False to remove this index

                #drop the action_prob
                output = output[:, :4]

                # Apply the mask to remove finished sequences
                if not mask.any():
                    break  # Exit the loop if all sequences are finished
                
                if iteration >= max_iter:
                    break
                else:
                    print(f"Iteration: {iteration}") if config['test'] else None
                    iteration += 1

                # Update end_idx and tensors based on the mask
                end_idx = [end_idx[j] for j in range(len(end_idx)) if mask[j]]
                input_seq = input_seq[mask]
                output = output[mask]
                
                # Create the required input sequence for the next step
                if len(config['other_features'])>0:
                    # other_features = ['team','home_team','success','seconds','deltaX','deltaY','distance','dist2goal','angle2goal']
                    #get the previous timestep
                    prev_timestep = input_seq[:,-1,:] if len(input_seq.shape) == 3 else input_seq
                    #create a tensor with the same shape as the prev_timestep
                    append_tensor = torch.zeros((prev_timestep.shape[0],len(config['features'])+config['num_actions']-1))
                    #convert the prev_timestep to a numpy array
                    prev_timestep = prev_timestep.cpu().numpy() if device != "cpu" else prev_timestep.numpy()
                    #get the previous timestep's action
                    prev_action = prev_timestep[:,0]
                    #create a position dict for the position of the config['other_features'] in config['features']
                    position_dict = {feature:config['features'].index(feature)+config['num_actions']-1 for feature in config['other_features']}
                    #if action is 3, then update the 'team','home_team','success','seconds'
                    for i, output_i in enumerate(output):
                        if output_i[0] == 3:
                            action = output_i[0]
                            delta_T = 0
                            x = output_i[2]
                            y = output_i[3]
                            team_idx= position_dict.get('team')
                            team = prev_timestep[i,team_idx] if team_idx is not None else None
                            home_team_idx = position_dict.get('home_team')
                            home_team = prev_timestep[i,home_team_idx] if home_team_idx is not None else None
                            success_idx = position_dict.get('success')
                            success = 0 if success_idx is not None else None
                            seconds_idx = position_dict.get('seconds')
                            if seconds_idx is not None:
                                seconds = prev_timestep[i,position_dict['seconds']]
                                seconds = seconds*(max_dict["seconds"]-min_dict["seconds"]) + min_dict["seconds"]
                                seconds = seconds + np.exp(prev_timestep[i,1]*(max_dict["delta_T"]-min_dict["delta_T"]) + min_dict["delta_T"]) - 1e-6
                            deltaX_idx = position_dict.get('deltaX')
                            deltaX = 0 if deltaX_idx is not None else None
                            deltaY_idx = position_dict.get('deltaY')
                            deltaY = 0 if deltaY_idx is not None else None
                            distance_idx = position_dict.get('distance')
                            distance = 0 if distance_idx is not None else None
                            dist2goal_idx = position_dict.get('dist2goal')
                            dist2goal = 0 if dist2goal_idx is not None else None
                            angle2goal_idx = position_dict.get('angle2goal')
                            angle2goal = 0.5 if angle2goal_idx is not None else None
                            home_score_idx = position_dict.get('home_score')
                            home_score = prev_timestep[i,position_dict['home_score']] if home_score_idx is not None else None
                            away_score_idx = position_dict.get('away_score')
                            away_score = prev_timestep[i,position_dict['away_score']] if away_score_idx is not None else None                        
                        else:
                            action = output_i[0]
                            delta_T = output_i[1]
                            x = output_i[2]
                            y = output_i[3]
                            team_idx= position_dict.get('team')
                            if team_idx is not None:
                                previous_team = prev_timestep[i,team_idx]
                                match_id = data.loc[end_idx[i]]['match_id']
                                other_team = data[data['match_id'] == match_id].team.unique().tolist()
                                other_team.remove(previous_team)
                                other_team = other_team[0]
                                team = previous_team if prev_action[i] !=3 else other_team
                                if not isinstance(team, int):
                                    team = team.astype(int)
                            home_team_idx = position_dict.get('home_team')
                            if home_team_idx is not None:
                                previous_home_team = prev_timestep[i,home_team_idx]
                                home_team = previous_home_team if prev_action[i] !=3 else abs(1-previous_home_team)
                                if not isinstance(home_team, int):
                                    home_team = home_team.astype(int)
                            success_idx = position_dict.get('success')
                            if success_idx is not None:
                                success = 1 if action not in [3,6] else 0
                                if not isinstance(success, int):
                                    success = success.astype(int)
                            seconds_idx = position_dict.get('seconds')
                            if seconds_idx is not None:
                                seconds = prev_timestep[i,position_dict['seconds']]
                                seconds = seconds*(max_dict["seconds"]-min_dict["seconds"]) + min_dict["seconds"]
                                seconds = seconds + np.exp(prev_timestep[i,1]*(max_dict["delta_T"]-min_dict["delta_T"]) + min_dict["delta_T"]) - 1e-6
                                if not isinstance(seconds, float):
                                    seconds = seconds.astype(float)
                            deltaX_idx = position_dict.get('deltaX')
                            if deltaX_idx is not None:
                                deltaX = x*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"]
                                deltaX = x - prev_timestep[i,2]*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"]
                            deltaY_idx = position_dict.get('deltaY')
                            if deltaY_idx is not None:
                                deltaY = y*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"]
                                deltaY = y - prev_timestep[i,3]*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"]
                            distance_idx = position_dict.get('distance')
                            if distance_idx is not None:
                                distance = torch.sqrt((deltaX)**2 + (deltaY)**2)
                            dist2goal_idx = position_dict.get('dist2goal')
                            if dist2goal_idx is not None:
                                dist2goal = torch.sqrt((x*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"] - 105)**2 + (y*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"] - 34)**2)
                            angle2goal_idx = position_dict.get('angle2goal')
                            if angle2goal_idx is not None:
                                angle2goal = torch.abs(torch.atan2((y*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"] - 34), (x*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"] - 105)))
                            home_score_idx = position_dict.get('home_score')
                            if home_score_idx is not None:
                                if action == 8 and home_team == 1:
                                    home_score = prev_timestep[i,position_dict['home_score']]+1
                                else:
                                    home_score = prev_timestep[i,position_dict['home_score']]
                            away_score_idx = position_dict.get('away_score')
                            if away_score_idx is not None:
                                if action == 8 and home_team == 0:
                                    away_score = prev_timestep[i,position_dict['away_score']]+1
                                else:
                                    away_score = prev_timestep[i,position_dict['away_score']]                         
                        #preprocess the features
                        for feature in config['other_features']:
                            if feature == 'seconds':
                                seconds = (seconds - min_dict['seconds'])/(max_dict['seconds'] - min_dict['seconds'])
                            elif feature == 'deltaX':
                                deltaX = (deltaX - min_dict['deltaX'])/(max_dict['deltaX'] - min_dict['deltaX'])
                            elif feature == 'deltaY':
                                deltaY = (deltaY - min_dict['deltaY'])/(max_dict['deltaY'] - min_dict['deltaY'])
                            elif feature == 'distance':
                                distance = (distance - min_dict['distance'])/(max_dict['distance'] - min_dict['distance'])
                            elif feature == 'dist2goal':
                                dist2goal = (dist2goal - min_dict['dist2goal'])/(max_dict['dist2goal'] - min_dict['dist2goal'])
                            elif feature == 'angle2goal':
                                angle2goal = (angle2goal - min_dict['angle2goal'])/(max_dict['angle2goal'] - min_dict['angle2goal'])
                            elif feature == 'home_score':
                                if max_dict['home_score'] - min_dict['home_score'] != 0:
                                    home_score = (home_score - min_dict['home_score'])/(max_dict['home_score'] - min_dict['home_score'])
                                else:
                                    home_score = 0
                            elif feature == 'away_score':
                                if max_dict['away_score'] - min_dict['away_score'] != 0:
                                    away_score = (away_score - min_dict['away_score'])/(max_dict['away_score'] - min_dict['away_score'])
                                else:
                                    away_score = 0                        
                        #update the append_tensor
                        for j in range(config['num_actions']):
                            if j == action:
                                append_tensor[i,j] = 1
                            else:
                                append_tensor[i,j] = 0
                        append_tensor[i,config['num_actions']] = delta_T
                        append_tensor[i,config['num_actions']+1] = x
                        append_tensor[i,config['num_actions']+2] = y
                        if team_idx is not None:
                            append_tensor[i,team_idx] = team
                        if home_team_idx is not None:
                            append_tensor[i,home_team_idx] = home_team.astype(int)
                        if success_idx is not None: #TODO: adjust this if success is added as a target feature
                            append_tensor[i,success_idx] = success
                        if seconds_idx is not None:
                            append_tensor[i,seconds_idx] = seconds.astype(float)
                        if deltaX_idx is not None:
                            append_tensor[i,deltaX_idx] = deltaX
                        if deltaY_idx is not None:
                            append_tensor[i,deltaY_idx] = deltaY
                        if distance_idx is not None:
                            append_tensor[i,distance_idx] = distance
                        if dist2goal_idx is not None:
                            append_tensor[i,dist2goal_idx] = dist2goal
                        if angle2goal_idx is not None:
                            append_tensor[i,angle2goal_idx] = angle2goal
                        if home_score_idx is not None:
                            append_tensor[i,home_score_idx] = home_score
                        if away_score_idx is not None:
                            append_tensor[i,away_score_idx] = away_score                    
                    append_tensor = append_tensor.to(device).unsqueeze(1)
                    if config['seq_len']==1:
                        input_seq = append_tensor.flatten(1)
                    else:
                        input_seq = torch.cat((input_seq[:, len(config['features'])-1+config['num_actions']:], append_tensor.flatten(1)), dim=-1) 
                else:
                    # Update the input sequence for the next step
                    if config['seq_len']==1:
                        input_seq = append_tensor.flatten(1)
                    else:
                        input_seq = torch.cat((input_seq[:, len(config['features'])-1+config['num_actions']:], append_tensor.flatten(1)), dim=-1)     

        rows = []
        for key, value_list in simulation.items():
            for value in value_list:
                rows.append([key] + value)

        # Convert the list of rows into a DataFrame
        df = pd.DataFrame(rows, columns=['index', 'iteration', 'action', 'action_prob', 'delta_T', 'delta_T_prob', 'x', 'x_prob', 'y', 'y_prob'])
        #set to 4 dp
        df = df.round(4)

        # Ensure for each index the last action is action=3
        for idx in df['index'].unique():
            # Get the last row for the given index
            last_row_idx = df[df['index'] == idx].index[-1]
            
            # Check if the last action is not 3
            if df.loc[last_row_idx, 'action'] != 3:
                # Change the last row's action to 3
                df.loc[last_row_idx, 'action'] = 3

    return df

def LEM_simulation_match(train_df, data, model_name, model_path, model_config, num_workers=4, batch_size=64, device=None, random_selection=False, max_iter=200, min_max_dict_path=None, simulation_time=45,match_id=None): 
        # Load the training configuration 
    with open(model_config, "r") as f:
        config = json.load(f)
    
    #check if train_df and data are paths or dataframes
    if train_df is not None and train_df != "None":
        if not isinstance(train_df, pd.DataFrame):
            #check if the train_df is a path
            if isinstance(train_df, str):
                train_df = pd.read_csv(train_df) if not config['test'] else pd.read_csv(train_df).head(1000)
            else:
                raise ValueError("train_df must be a pandas dataframe or a path to a csv file")
    if isinstance(data, pd.DataFrame):
        data_raw = data
    elif isinstance(data, str):
        data_raw = pd.read_csv(data) if not config['test'] else pd.read_csv(data).head(1000)
    else:
        raise ValueError("data must be a pandas dataframe or a path to a csv file")
    data = data_raw.copy()

    #set the hyperparameters
    seq_len = config["seq_len"]
    features = config["features"]
    num_workers = config["dataloader_num_worker"]
    batch_size = config["batch_size"]
    device = config["device"]
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device
    
    # Preprocess the data
    if min_max_dict_path != "None" and min_max_dict_path is not None:
        try:
            with open(min_max_dict_path, "r") as f:
                min_max_dict = json.load(f)
            min_dict = min_max_dict['min_dict']
            max_dict = min_max_dict['max_dict']
        except:
            train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    else:
        train_df, min_dict, max_dict = UEID_preprocessing(train_df)

    data, _, _ = UEID_preprocessing(data, min_dict, max_dict)

    #for each match, run the simulation for simulation_time and only initialize the simulation for the first row with the data
    if match_id is not None:
        match_ids = [match_id]
    else:
        match_ids = [data['match_id'].unique()[0]]
    match_simulation_data = []
    simulation_period = 1 if simulation_time >= 45 else 2
    simulation_minute = 90 - simulation_time 
    for match_id in match_ids:
        match_data = data[data['match_id'] == match_id]
        team_list = match_data['team'].unique().tolist()
        #find the first row where the column 'Period' is equal to simulation_period and 'Minute' is equal to simulation_minute
        try:
            start_idx = match_data[(match_data['Period'] == simulation_period) & (match_data['Minute'] == simulation_minute)].index[0]
        except:
            start_idx = match_data[(match_data['Period'] == simulation_period) & (match_data['Minute'] == simulation_minute+1)].index[0]
        #append the first and the following seq_len-1 rows to the match_simulation_data
        match_simulation_data.append(match_data.iloc[start_idx:start_idx+seq_len+1])

    #convert the match_simulation_data to a dataframe
    match_simulation_data = pd.concat(match_simulation_data)
    # Create the data loader
    data_loader = UEID_Simulation_loader(match_simulation_data, seq_len, features)
    data_loader = DataLoader(data_loader, batch_size=batch_size, num_workers=num_workers, shuffle=False, drop_last=False)

    # Load the model
    model = load_model(model_name, model_path, model_config)
    model.load_state_dict(torch.load(model_path, weights_only=True))
    model.to(device)
    model.eval()

    # Run inference
    with torch.no_grad():
        simulation = {}
        for i, (input_seq, end_idx) in tqdm(enumerate(data_loader)):
            #convert the end_idx to a list
            end_idx = end_idx.cpu().numpy().tolist() if device != "cpu" else end_idx.numpy().tolist()
            for idx in end_idx:
                simulation[idx] = []
            num_action=config["num_actions"] if config is not None and config !="None" else 9
            inputs_one_hot = torch.nn.functional.one_hot(input_seq[:,:,0].long(), num_classes=num_action).float()
            input_seq = torch.cat((inputs_one_hot, input_seq[:,:,1:]), dim=2)
            batch_size = input_seq.shape[0]
            input_seq = input_seq.reshape(batch_size, -1)
            input_seq = input_seq.to(device)
            iteration = 1
            time = [0]*len(end_idx)
            # Run the simulation
            while end_idx:

                # Forward pass through the model
                output = model(input_seq)

                # Convert the first config['num_actions'] to action probabilities
                action_probs = torch.nn.functional.softmax(output[:, :config['num_actions']], dim=1)
                delta_T_probs = torch.nn.functional.softmax(output[:, config['num_actions']:config['num_actions']+config['delta_T_bin']], dim=1)
                start_x_probs = torch.nn.functional.softmax(output[:, config['num_actions']+config['delta_T_bin']:config['num_actions']+config['delta_T_bin']+config['start_x_bin']], dim=1)
                start_y_probs = torch.nn.functional.softmax(output[:, config['num_actions']+config['delta_T_bin']+config['start_x_bin']:], dim=1)

                # Select actions based on the probabilities
                if random_selection:
                    action = torch.multinomial(action_probs, 1)
                    delta_T = torch.multinomial(delta_T_probs, 1)
                    start_x = torch.multinomial(start_x_probs, 1)
                    start_y = torch.multinomial(start_y_probs, 1)
                    action_prob = torch.gather(action_probs, 1, action)
                    delta_T_prob = torch.gather(delta_T_probs, 1, delta_T)
                    start_x_prob = torch.gather(start_x_probs, 1, start_x)
                    start_y_prob = torch.gather(start_y_probs, 1, start_y)
                    delta_T = delta_T/100
                    start_x = start_x/100
                    start_y = start_y/100
                else:
                    action = torch.argmax(action_probs, dim=1)
                    action = action.unsqueeze(1)
                    action_prob = torch.gather(action_probs, 1, action)
                    delta_T = torch.argmax(delta_T_probs, dim=1)
                    delta_T = delta_T.unsqueeze(1)
                    delta_T_prob = torch.gather(delta_T_probs, 1, delta_T)
                    delta_T = delta_T/100
                    start_x = torch.argmax(start_x_probs, dim=1)
                    start_x = start_x.unsqueeze(1)
                    start_x_prob = torch.gather(start_x_probs, 1, start_x)
                    start_x = start_x/100
                    start_y = torch.argmax(start_y_probs, dim=1)
                    start_y = start_y.unsqueeze(1)
                    start_y_prob = torch.gather(start_y_probs, 1, start_y)
                    start_y = start_y/100
                    
                # Recombine the action with the delta_T, start_x, start_y
                
                output = torch.cat((action.float(), delta_T.float(), start_x.float(), start_y.float(), action_prob.float(), delta_T_prob.float(), start_x_prob.float(), start_y_prob.float()), dim=1)

                # Update the simulation dictionary
                mask = torch.ones(len(end_idx), dtype=bool)  # Initialize a mask to keep all entries

                for j, idx in enumerate(end_idx):
                    output_j = output[j].cpu().numpy().tolist() if device != "cpu" else output[j].numpy().tolist()
                    action_j = int(output_j[0])
                    delta_T_j = np.exp(output_j[1]*(max_dict["delta_T"]-min_dict["delta_T"]) + min_dict["delta_T"]) - 1e-6
                    start_x_j = output_j[2]*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"]
                    start_y_j = output_j[3]*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"]
                    
                    action_prob_j = output_j[4]
                    delta_T_prob_j = output_j[5]
                    start_x_prob_j = output_j[6]
                    start_y_prob_j = output_j[7]
                    if 'home_score' in config['features'] and 'away_score' in config['features']:
                        home_score_idx = config['features'].index('home_score')-1+config['num_actions']+(len(config['features'])-1+config['num_actions'])*(seq_len-1)
                        away_score_idx = config['features'].index('away_score')-1+config['num_actions']+(len(config['features'])-1+config['num_actions'])*(seq_len-1)
                        home_score_j = input_seq[j][home_score_idx]
                        away_score_j = input_seq[j][away_score_idx]
                        #detach the home_score and away_score
                        home_score_j = home_score_j.cpu().numpy().tolist() if device != "cpu" else home_score_j.numpy().tolist()
                        away_score_j = away_score_j.cpu().numpy().tolist() if device != "cpu" else away_score_j.numpy().tolist()
                        home_score_j = int(home_score_j)
                        away_score_j = int(away_score_j)
                        #unnormalize the home_score and away_score
                        if max_dict['home_score'] - min_dict['home_score'] != 0:
                            home_score_j = home_score_j*(max_dict["home_score"]-min_dict["home_score"]) + min_dict["home_score"]
                            away_score_j = away_score_j*(max_dict["away_score"]-min_dict["away_score"]) + min_dict["away_score"]
                            
                    #check if idx is a key in the simulation dictionary
                    if idx not in simulation.keys():
                        simulation[idx] = []
                    simulation[idx].append([action_j, action_prob_j, delta_T_j, delta_T_prob_j, start_x_j, start_x_prob_j, start_y_j, start_y_prob_j]+[home_score_j, away_score_j])
                    time[j] += delta_T_j
                    print(f"Time: {time[j]}") if config['test'] else None
                    if time[j] >= 60 and config['test']:
                        mask[j] = False

                    if time[j] >= simulation_time*60:
                        mask[j] = False  # Mark as False to remove this index

                #drop the action_prob
                output = output[:, :4]

                # Apply the mask to remove finished sequences
                if not mask.any():
                    break  # Exit the loop if all sequences are finished
                
                if iteration >= max_iter:
                    output_i[0]=3
                iteration += 1
                print(f"Iteration: {iteration}") if config['test'] else None

                # # Update end_idx and tensors based on the mask
                # end_idx = [end_idx[j] for j in range(len(end_idx)) if mask[j]]
                # input_seq = input_seq[mask]
                # output = output[mask]
                
                # Create the required input sequence for the next step
                if len(config['other_features'])>0:
                    # other_features = ['team','home_team','success','seconds','deltaX','deltaY','distance','dist2goal','angle2goal']
                    #get the previous timestep
                    prev_timestep = input_seq[:,seq_len:] if len(input_seq.shape) == 3 else input_seq
                    #create a tensor with the same shape as the prev_timestep
                    append_tensor = torch.zeros((prev_timestep.shape[0],len(config['features'])+config['num_actions']-1))
                    #convert the prev_timestep to a numpy array
                    prev_timestep = prev_timestep.cpu().numpy() if device != "cpu" else prev_timestep.numpy()
                    #get the previous timestep's action
                    prev_action = prev_timestep[:,0]
                    #create a position dict for the position of the config['other_features'] in config['features']
                    position_dict = {feature:config['features'].index(feature)+config['num_actions']-1 for feature in config['other_features']}
                    #if action is 3, then update the 'team','home_team','success','seconds'
                    for i, output_i in enumerate(output):
                        if output_i[0] == 3:
                            end_idx[i] = end_idx[i] + 1
                            action = output_i[0]
                            delta_T = 0
                            x = output_i[2]
                            y = output_i[3]
                            team_idx= position_dict.get('team')
                            team = prev_timestep[i,team_idx] if team_idx is not None else None
                            home_team_idx = position_dict.get('home_team')
                            home_team = prev_timestep[i,home_team_idx] if home_team_idx is not None else None
                            success_idx = position_dict.get('success')
                            success = 0 if success_idx is not None else None
                            seconds_idx = position_dict.get('seconds')
                            if seconds_idx is not None:
                                seconds = prev_timestep[i,position_dict['seconds']]
                                seconds = seconds*(max_dict["seconds"]-min_dict["seconds"]) + min_dict["seconds"]
                                seconds = seconds + np.exp(prev_timestep[i,1]*(max_dict["delta_T"]-min_dict["delta_T"]) + min_dict["delta_T"]) - 1e-6
                            deltaX_idx = position_dict.get('deltaX')
                            deltaX = 0 if deltaX_idx is not None else None
                            deltaY_idx = position_dict.get('deltaY')
                            deltaY = 0 if deltaY_idx is not None else None
                            distance_idx = position_dict.get('distance')
                            distance = 0 if distance_idx is not None else None
                            dist2goal_idx = position_dict.get('dist2goal')
                            dist2goal = 0 if dist2goal_idx is not None else None
                            angle2goal_idx = position_dict.get('angle2goal')
                            angle2goal = 0.5 if angle2goal_idx is not None else None
                            home_score_idx = position_dict.get('home_score')
                            home_score = prev_timestep[i,position_dict['home_score']] if home_score_idx is not None else None
                            away_score_idx = position_dict.get('away_score')
                            away_score = prev_timestep[i,position_dict['away_score']] if away_score_idx is not None else None                        
                        else:
                            action = output_i[0]
                            delta_T = output_i[1]
                            x = output_i[2]
                            y = output_i[3]
                            team_idx= position_dict.get('team')
                            if team_idx is not None:
                                previous_team = prev_timestep[i,team_idx]
                                match_id = data.loc[end_idx[i]]['match_id']
                                other_team = data[data['match_id'] == match_id].team.unique().tolist()
                                other_team.remove(previous_team)
                                other_team = other_team[0]
                                team = previous_team if prev_action[i] !=3 else other_team
                                if not isinstance(team, int):
                                    team = team.astype(int)
                            home_team_idx = position_dict.get('home_team')
                            if home_team_idx is not None:
                                previous_home_team = prev_timestep[i,home_team_idx]
                                home_team = previous_home_team if prev_action[i] !=3 else abs(1-previous_home_team)
                                if not isinstance(home_team, int):
                                    home_team = home_team.astype(int)
                            success_idx = position_dict.get('success')
                            if success_idx is not None:
                                success = 1 if action not in [3,6] else 0
                                if not isinstance(success, int):
                                    success = success.astype(int)
                            seconds_idx = position_dict.get('seconds')
                            if seconds_idx is not None:
                                seconds = prev_timestep[i,position_dict['seconds']]
                                seconds = seconds*(max_dict["seconds"]-min_dict["seconds"]) + min_dict["seconds"]
                                seconds = seconds + np.exp(prev_timestep[i,1]*(max_dict["delta_T"]-min_dict["delta_T"]) + min_dict["delta_T"]) - 1e-6
                                if not isinstance(seconds, float):
                                    seconds = seconds.astype(float)
                            deltaX_idx = position_dict.get('deltaX')
                            if deltaX_idx is not None:
                                deltaX = x*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"]
                                deltaX = x - prev_timestep[i,2]*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"]
                            deltaY_idx = position_dict.get('deltaY')
                            if deltaY_idx is not None:
                                deltaY = y*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"]
                                deltaY = y - prev_timestep[i,3]*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"]
                            distance_idx = position_dict.get('distance')
                            if distance_idx is not None:
                                distance = torch.sqrt((deltaX)**2 + (deltaY)**2)
                            dist2goal_idx = position_dict.get('dist2goal')
                            if dist2goal_idx is not None:
                                dist2goal = torch.sqrt((x*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"] - 105)**2 + (y*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"] - 34)**2)
                            angle2goal_idx = position_dict.get('angle2goal')
                            if angle2goal_idx is not None:
                                angle2goal = torch.abs(torch.atan2((y*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"] - 34), (x*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"] - 105)))
                            home_score_idx = position_dict.get('home_score')
                            if home_score_idx is not None:
                                if action == 8 and home_team == 1:
                                    home_score = prev_timestep[i,position_dict['home_score']]+1
                                else:
                                    home_score = prev_timestep[i,position_dict['home_score']]
                            away_score_idx = position_dict.get('away_score')
                            if away_score_idx is not None:
                                if action == 8 and home_team == 0:
                                    away_score = prev_timestep[i,position_dict['away_score']]+1
                                else:
                                    away_score = prev_timestep[i,position_dict['away_score']]                        
                        #preprocess the features
                        for feature in config['other_features']:
                            if feature == 'seconds':
                                seconds = (seconds - min_dict['seconds'])/(max_dict['seconds'] - min_dict['seconds'])
                            elif feature == 'deltaX':
                                deltaX = (deltaX - min_dict['deltaX'])/(max_dict['deltaX'] - min_dict['deltaX'])
                            elif feature == 'deltaY':
                                deltaY = (deltaY - min_dict['deltaY'])/(max_dict['deltaY'] - min_dict['deltaY'])
                            elif feature == 'distance':
                                distance = (distance - min_dict['distance'])/(max_dict['distance'] - min_dict['distance'])
                            elif feature == 'dist2goal':
                                dist2goal = (dist2goal - min_dict['dist2goal'])/(max_dict['dist2goal'] - min_dict['dist2goal'])
                            elif feature == 'angle2goal':
                                angle2goal = (angle2goal - min_dict['angle2goal'])/(max_dict['angle2goal'] - min_dict['angle2goal'])
                            elif feature == 'home_score':
                                if max_dict['home_score'] - min_dict['home_score'] != 0:
                                    home_score = (home_score - min_dict['home_score'])/(max_dict['home_score'] - min_dict['home_score'])
                                else:
                                    home_score = 0
                            elif feature == 'away_score':
                                if max_dict['away_score'] - min_dict['away_score'] != 0:
                                    away_score = (away_score - min_dict['away_score'])/(max_dict['away_score'] - min_dict['away_score'])
                                else:
                                    away_score = 0                        
                        #update the append_tensor
                        for j in range(config['num_actions']):
                            if j == action:
                                append_tensor[i,j] = 1
                            else:
                                append_tensor[i,j] = 0
                        append_tensor[i,config['num_actions']] = delta_T
                        append_tensor[i,config['num_actions']+1] = x
                        append_tensor[i,config['num_actions']+2] = y
                        if team_idx is not None:
                            append_tensor[i,team_idx] = int(team)
                        if home_team_idx is not None:
                            append_tensor[i,home_team_idx] = home_team.astype(int)
                        if success_idx is not None: #TODO: adjust this if success is added as a target feature
                            append_tensor[i,success_idx] = success
                        if seconds_idx is not None:
                            append_tensor[i,seconds_idx] = seconds.astype(float)
                        if deltaX_idx is not None:
                            append_tensor[i,deltaX_idx] = deltaX
                        if deltaY_idx is not None:
                            append_tensor[i,deltaY_idx] = deltaY
                        if distance_idx is not None:
                            append_tensor[i,distance_idx] = distance
                        if dist2goal_idx is not None:
                            append_tensor[i,dist2goal_idx] = dist2goal
                        if angle2goal_idx is not None:
                            append_tensor[i,angle2goal_idx] = angle2goal
                        if home_score_idx is not None:
                            append_tensor[i,home_score_idx] = home_score
                        if away_score_idx is not None:
                            append_tensor[i,away_score_idx] = away_score                    
                    append_tensor = append_tensor.to(device).unsqueeze(1)
                    if config['seq_len']==1:
                        input_seq = append_tensor.flatten(1)
                    else:
                        input_seq = torch.cat((input_seq[:, len(config['features'])-1+config['num_actions']:], append_tensor.flatten(1)), dim=-1) 
                else:
                    # Update the input sequence for the next step
                    if config['seq_len']==1:
                        input_seq = append_tensor.flatten(1)
                    else:
                        input_seq = torch.cat((input_seq[:, len(config['features'])-1+config['num_actions']:], append_tensor.flatten(1)), dim=-1)     

        rows = []
        for key, value_list in simulation.items():
            for value in value_list:
                rows.append([key] + value)

        # Convert the list of rows into a DataFrame
        columns = ['index', 'action', 'action_prob', 'delta_T', 'delta_T_prob', 'x', 'x_prob', 'y', 'y_prob']
        columns += ['home_score', 'away_score'] if 'home_score' in config['features'] and 'away_score' in config['features'] else []
        df = pd.DataFrame(rows, columns=columns)
        #set to 4 dp
        df = df.round(4)

        # Ensure for each index the last action is action=3
        for idx in df['index'].unique():
            # Get the last row for the given index
            last_row_idx = df[df['index'] == idx].index[-1]
            
            # Check if the last action is not 3
            if df.loc[last_row_idx, 'action'] != 3:
                # Change the last row's action to 3
                df.loc[last_row_idx, 'action'] = 3

    return df

def load_model(model_name:str,model_path:str,model_config:str):
    if model_name == "LEM":
        from ..models.LEM import LEM
        with open(model_config, "r") as f:
            config = json.load(f)
        hidden_size = config["hidden_size"]
        output_size = config["output_size"]
        input_size_action = config["input_size_action"]
        hidden_size_action = config["hidden_size_action"]
        output_size_action = config["output_size_action"]
        activation = config["activation"]
        activation_action = config["activation_action"]
        pth_action = config["LEM_aciton_pth_path"]
        model = LEM(hidden_size, output_size, input_size_action, hidden_size_action, output_size_action,pth_action,
            activation,activation_action)
        model.load_state_dict(torch.load(model_path, weights_only=True))
        return model
    else:
        raise ValueError(f"Model {model_name} not supported")

def simulation_evaluation(simulation_df, ground_truth_df):
    #check if simulation_df and ground_truth_df are paths or dataframes
    if not isinstance(simulation_df, pd.DataFrame):
        #check if the simulation_df is a path
        if isinstance(simulation_df, str):
            simulation_df = pd.read_csv(simulation_df)
        else:
            raise ValueError("simulation_df must be a pandas dataframe or a path to a csv file")
    if not isinstance(ground_truth_df, pd.DataFrame):
        #check if the ground_truth_df is a path
        if isinstance(ground_truth_df, str):
            ground_truth_df = pd.read_csv(ground_truth_df)
        else:
            raise ValueError("ground_truth_df must be a pandas dataframe or a path to a csv file")
    
    #drop the game_over and period_over rows
    ground_truth_df = ground_truth_df[ground_truth_df["action"] != "game_over"]
    ground_truth_df = ground_truth_df[ground_truth_df["action"] != "period_over"]

    #get the unique indexes
    unique_indexes = simulation_df['index'].unique()
    #evaluating per timestep
    error_row = []
    timestep_eval = {}
    es_hota_list= []
    action_dict = {'short_pass': 0, 'carry': 1, 'high_pass': 2, '_': 3, 'cross': 4, 
                   'long_pass': 5, 'shot': 6, 'dribble': 7}
    #map the action to the action_dict
    ground_truth_df['action'] = ground_truth_df['action'].map(action_dict)
    ground_truth_df.loc[(ground_truth_df["action"] == 6) & (ground_truth_df["goal"] == 1), "action"] = 8

    for idx in tqdm(unique_indexes):
        try:
            sim_poss = simulation_df[simulation_df['index'] == idx]
            #get the ground truth (the index and the following action with the same poss_id)
            ground_truth = ground_truth_df.loc[idx]
            poss_id = ground_truth['poss_id']
            match_id = ground_truth['match_id']
            ground_truth_possession = ground_truth_df[(ground_truth_df['poss_id'] == poss_id) & (ground_truth_df['match_id'] == match_id)]
            #drop the row with index less than the indx
            ground_truth_possession = ground_truth_possession[ground_truth_possession.index >= idx]
            
            #evaluate the simulation per timestep
            for step_i, (_, row) in enumerate(sim_poss.iterrows()):
                #check if the simulation timestep exceed the ground truth timestep
                if step_i >= len(ground_truth_possession):
                    #next idx
                    break
                #initialize the row_eval if the timestep is not in the row_eval
                if step_i not in timestep_eval:
                    timestep_eval[step_i] = {"ACC_action":[],"MAE_delta_T":[],"MAE_start_x":[],"MAE_start_y":[]} 
                #get the ground truth for the current timestep
                ACC_action = 1 if int(row['action']) == int(ground_truth_possession.iloc[step_i]['action']) else 0
                MAE_delta_T = abs(row['delta_T'] - ground_truth_possession.iloc[step_i]['delta_T'])
                MAE_start_x = abs(row['x'] - ground_truth_possession.iloc[step_i]['start_x'])
                MAE_start_y = abs(row['y'] - ground_truth_possession.iloc[step_i]['start_y'])
                timestep_eval[step_i]["ACC_action"].append(ACC_action)
                timestep_eval[step_i]["MAE_delta_T"].append(MAE_delta_T)
                timestep_eval[step_i]["MAE_start_x"].append(MAE_start_x)
                timestep_eval[step_i]["MAE_start_y"].append(MAE_start_y)
                    
            #ES-HOTA
            es_hota_list.append([idx]+ES_HOTA_cal(sim_poss,ground_truth_possession))

        except:
            error_row.append([idx])
            continue

    #convert the timestep_eval to a dataframe with key as the column time_step and the loss as other columns
    time_step_data = []
    for time_step, metrics in timestep_eval.items():
        mean_entry = {
            'time_step': time_step,
            'ACC_action': sum(metrics['ACC_action']) / len(metrics['ACC_action']),
            'MAE_delta_T': sum(metrics['MAE_delta_T']) / len(metrics['MAE_delta_T']),
            'MAE_start_x': sum(metrics['MAE_start_x']) / len(metrics['MAE_start_x']),
            'MAE_start_y': sum(metrics['MAE_start_y']) / len(metrics['MAE_start_y'])
        }
        time_step_data.append(mean_entry)

    # Create DataFrame
    timestep_eval_df = pd.DataFrame(time_step_data)
    #add a row for the average of the timestep_eval
    timestep_eval_df.loc[len(timestep_eval_df)] = ['Overall']+ timestep_eval_df.iloc[:,1:].mean().tolist()

    #round the values to 4 decimal places
    timestep_eval_df = timestep_eval_df.round(4)
    es_hota_df = pd.DataFrame(es_hota_list, columns=['idx','TP','FN','FP']+[f"ES_HOTA_{i:.2f}" for i in np.arange(0.05,1,0.05)]+['ES_HOTA'])
    #add a row for the average of the ES_HOTA
    es_hota_df.loc[len(es_hota_df)] = ['Overall']+ es_hota_df.iloc[:,1:4].sum().tolist()+ es_hota_df.iloc[:,4:-1].mean().tolist() + [es_hota_df.iloc[:,-1].mean()]
    es_hota_df = es_hota_df.round(4)

    #print the error rows
    if error_row:
        print(f"Error idx: {error_row}")
    #print the es_hota mean value and es_hota 0.5 value
    print('Results for the Simulation Evaluation:')
    print(f"ES-HOTA: {es_hota_df.iloc[-1,-1]:.4f} | ES-HOTA_0.5: {es_hota_df.iloc[-1,14]:.4f} | TP: {es_hota_df.iloc[-1,1]} | FN: {es_hota_df.iloc[-1,2]} | FP: {es_hota_df.iloc[-1,3]}")

    return timestep_eval_df, es_hota_df

def ES_HOTA_cal(sim_poss,ground_truth_possession,tau_t=5,tau_l=5):
    sim_list = []
    ES_HOTA_list = []
    for i in range(max(len(sim_poss),len(ground_truth_possession))):
        if i < len(ground_truth_possession) and i < len(sim_poss):
            IT_sim = np.exp(np.log(0.05)*(sim_poss.iloc[i]['delta_T']-ground_truth_possession.iloc[i]['delta_T'])**2/tau_t**2)
            dist = (sim_poss.iloc[i]['x']-ground_truth_possession.iloc[i]['start_x'])**2 + (sim_poss.iloc[i]['y']-ground_truth_possession.iloc[i]['start_y'])**2
            LOC_sim = np.exp(np.log(0.05)*dist/tau_l**2)
            ACT_sim = sim_poss.iloc[i]['action_prob'] if int(sim_poss.iloc[i]['action']) == int(ground_truth_possession.iloc[i]['action']) else 0
            sim_score=IT_sim*LOC_sim*ACT_sim
            sim_list.append(sim_score)
        else:
            sim_list.append(0)

    for sim_i in np.arange(0.05,1,0.05):
        #convert the sim_list to binary values where the value is 1 if the value is greater than sim_i
        sim_list_binary = [1 if sim >= sim_i else 0 for sim in sim_list]
        len_sim = len(sim_list)
        len_gt = len(ground_truth_possession)
        TP_count = sum(sim_list_binary)
        FP_count = len_sim - len_gt if len_sim > len_gt else 0
        #sum the gt_list_binary for 0 values
        FN_count = len_gt - TP_count

        A_count = 0
        for j in range(TP_count):
            # count sim_list_binary[:TP_count+1] for the element equal to sim_list_binary(j)
            TPA_count = TP_count 
            FNA_count = len_gt - TP_count
            FPA_count = len_sim - len_gt if len_sim > len_gt else 0
            A_count += TPA_count/(TPA_count+FNA_count+FPA_count)
        
        ES_HOTA_list.append(np.sqrt(A_count/(TP_count+FN_count+FP_count)))
    
    #get the average of the ES_HOTA_list
    ES_HOTA_value = sum(ES_HOTA_list)/len(ES_HOTA_list)
    ES_HOTA_list.append(ES_HOTA_value)

    #add the TP, FN, FP at the beginning of the list
    ES_HOTA_list = [TP_count,FN_count,FP_count] + ES_HOTA_list

    return ES_HOTA_list

if __name__ == "__main__":
    model_path = os.getcwd()+"/test/model/LEM/out/optuna/20241009_010549/run_1/_model_1.pth"
    model_config = os.getcwd()+"/test/model/LEM/out/optuna/20241009_010549/run_1/hyperparameters.json"
    min_max_dict_path = os.getcwd()+"/test/model/LEM/out/optuna/20241009_010549/min_max_dict.json"

    train_path = os.getcwd()+"/test/dataset/csv/train.csv"
    valid_path = os.getcwd()+"/test/dataset/csv/valid.csv"
    save_path = os.getcwd()+"/test/inference/LEM/"
    os.makedirs(save_path, exist_ok=True)

    #testing inference for LEM
    # inferenced_data, loss_df = LEM_inference(train_path, valid_path, "LEM", model_path, model_config)
    # inferenced_data, loss_df = LEM_inference(None, valid_path, "LEM", model_path, model_config, min_max_dict_path=min_max_dict_path)
    # inferenced_data.to_csv(save_path+"inference.csv",index=False)
    # if loss_df is not None:
    #     loss_df.to_csv(save_path+"loss.csv",index=False)
    # df = LEM_simulation_possession(train_path, valid_path, "LEM", model_path, model_config, random_selection=True, max_iter=20)
    # df.to_csv(save_path+"simulation.csv",index=False)
    df_90 = LEM_simulation_match(train_path, valid_path, "LEM", model_path, model_config, random_selection=True, max_iter=20, min_max_dict_path=min_max_dict_path, simulation_time=90)
    df_90.to_csv(save_path+"simulation_90.csv",index=False)

    # testing evaluation
    # simulation_df_path = os.getcwd()+"/test/inference/LEM/simulation.csv"
    # ground_truth_df_path = os.getcwd()+"/test/dataset/csv/valid.csv"
    # timestep_eval_df,es_hota_df = simulation_evaluation(simulation_df_path, ground_truth_df_path)
    # timestep_eval_df.to_csv(save_path+"timestep_eval.csv",index=False)
    # es_hota_df.to_csv(save_path+"ES_HOTA.csv",index=False)
    # pdb.set_trace()
    print('___________done______________')


import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.nn.init as init
import math
import numpy as np
import pandas as pd
import json
import os
import time

from ..dataloaders.UEID_loader import UEID_loader
from ..trainers.LEM_train import LEM_action_train
from ..utils.UEID_preprocessing import UEID_preprocessing

import pdb

class LEM_action(nn.Module): 
    #ref https://github.com/nvsclub/LargeEventsModel/blob/main/lib/model_utils.py#L94
    def __init__(self, input_size, hidden_size, output_size, activation='relu'):
        super(LEM_action, self).__init__()

        activation_dict = {
            'relu': nn.ReLU,
            'sigmoid': nn.Sigmoid,
            'tanh': nn.Tanh,
            'leaky_relu': nn.LeakyReLU,
        }
        layers = [
            nn.Linear(input_size, hidden_size[0]),
            activation_dict[activation]()
        ] + flatten([
            [nn.Linear(hidden_size[i], hidden_size[i+1]),
            activation_dict[activation]()] for i in range(len(hidden_size) - 1)
        ]) + [
            nn.Linear(hidden_size[-1], output_size),
            nn.Sigmoid()
        ]

        self.model = nn.Sequential(*layers)
        
        # Initialize the linear layers
        self.init_weights()

    def init_weights(self):
        for m in self.model.modules():
            if isinstance(m, nn.Linear):
                init.xavier_uniform_(m.weight)
                init.zeros_(m.bias)
    
    def forward(self, x):
        return self.model(x)

def flatten(l):
    return [item for sublist in l for item in sublist]

def train_LEM_action(current_time, config, train_df, valid_df, seq_len, features, epochs, batch_size,
            hidden_size, activation,
            num_actions=9, device=None, print_freq=10, patience=5, num_workers=4, optuna=False,
            learning_rate=0.01,eps=1e-16):

    # check if the train_df and valid_df are pandas dataframe
    if not isinstance(train_df, pd.DataFrame):
        #check if the train_df is a path
        if isinstance(train_df, str):
            train_df = pd.read_csv(train_df) if not config['test'] else pd.read_csv(train_df).head(1000)
        else:
            raise ValueError("train_df must be a pandas dataframe or a path to a csv file")
    if not isinstance(valid_df, pd.DataFrame):
        #check if the valid_df is a path
        if isinstance(valid_df, str):
            valid_df = pd.read_csv(valid_df) if not config['test'] else pd.read_csv(valid_df).head(1000)
        else:
            raise ValueError("valid_df must be a pandas dataframe or a path to a csv file")
        
    # Preprocess the data
    train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    valid_df, _, _ = UEID_preprocessing(valid_df, min_dict, max_dict)
    
    # Define fixed hyperparameters 
    input_size=(num_actions+len(features)-1)*seq_len
    output_size=num_actions

    #if not device is given, check if cuda is available
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device

    # Initialize model layers based on data
    model = LEM_action(input_size, hidden_size, output_size, activation).to(device)
    
    # Create data loaders
    train_loader = UEID_loader(train_df, seq_len, features)
    valid_loader = UEID_loader(valid_df, seq_len, features)

    # Convert data loaders to torch DataLoader
    train_loader = DataLoader(train_loader, batch_size=batch_size, shuffle=True, drop_last=True, num_workers=num_workers)
    valid_loader = DataLoader(valid_loader, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=num_workers)
    
    # Define optimizer
    optimizer = torch.optim.Adam(model.parameters(),lr=learning_rate,eps=eps)

    # Training logic here
    model_state_dict, best_epoch, train_losses, valid_losses, train_loss_components, valid_loss_components, model_stats = LEM_action_train(model, train_loader, valid_loader, optimizer, device, epochs, print_freq, patience,config=config)

    #create one csv file for the training and validation losses
    columns = ['train_loss', 'BCEL_action', 'ACC_action', 'f1_action',
                'valid_loss', 'BCEL_action_v', 'ACC_action_v', 'f1_action_v']
    data = [train_losses, np.array(train_loss_components)[:,0], np.array(train_loss_components)[:,1], np.array(train_loss_components)[:,2],
            valid_losses, np.array(valid_loss_components)[:,0], np.array(valid_loss_components)[:,1], np.array(valid_loss_components)[:,2]]
    data = np.array(data).T
    loss_df = pd.DataFrame(data, columns=columns)
    #round the loss_df to 4 decimal places
    loss_df = loss_df.round(4)

    #save the model and the loss_df
    if config['optuna']:
        method = "optuna"
    else:
        method = "train"
    i=1
    save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    while os.path.exists(save_path):
        i+=1
        save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    os.makedirs(save_path)

    model_save_path = save_path + f"_model_{best_epoch}.pth"
    loss_save_path = save_path + "_loss.csv"
    hyperparameters_save_path = save_path + "hyperparameters.json"
    model_stats_save_path = save_path + "_model_stats.txt"
    
    #save the min_dict and max_dict
    min_max_dict_path = config['save_path']+f"/out/{method}/{current_time}/min_max_dict.json"
    with open(min_max_dict_path, 'w') as f:
        json.dump({'min_dict':min_dict, 'max_dict':max_dict}, f, indent=4)

    #save all the hyperparameters values
    hyperparameters = {'current_time': current_time,
                        'train_path': config['train_path'], 'valid_path': config['valid_path'], 'save_path': config['save_path'],
                        'test': config['test'], 'batch_size': batch_size, 'num_epoch': config['num_epoch'],
                        'print_freq': print_freq, 'early_stop_patience': config['early_stop_patience'],
                        'dataloader_num_worker': config['dataloader_num_worker'], 'device': config['device'], 
                        'basic_features': config['basic_features'], 'other_features': config['other_features'],
                        'use_other_features': config['use_other_features'],
                        'features':features, 'num_actions': num_actions, 'seq_len': seq_len,
                        'optuna': config['optuna'], 'optuna_n_trials': config['optuna_n_trials'],
                        'learning_rate': learning_rate, 'eps': eps,
                        'input_size': input_size, 'hidden_size': hidden_size, 'output_size': output_size,
                        'activation': activation,
                        'model':model.__class__.__name__,
                        'best_epoch':best_epoch}
                
    torch.save(model_state_dict, model_save_path)
    loss_df.to_csv(loss_save_path, index=False)
    with open(hyperparameters_save_path, 'w') as f:
        json.dump(hyperparameters, f, indent=4)
    with open(model_stats_save_path, "w") as f:
        json.dump(model_stats, f, indent=4)
        
    print("Model and loss saved at", save_path)

    if optuna:
        return valid_losses[best_epoch]

def main(config):
    train_path = config['train_path']
    valid_path = config['valid_path']

    # Extract input dimensions from train_df
    if config['test']:
        seq_len = config['seq_len']
        epochs = 2
        features = config['basic_features']+config['other_features']
    else:
        seq_len = config['seq_len']
        epochs = config['num_epoch']
        if config['use_other_features']:
            features = config['basic_features']+config['other_features']
        else:
            features = config['basic_features']
    
    batch_size = config['batch_size']
    num_actions = config['num_actions']
    print_freq = config['print_freq']
    early_stop_patience = config['early_stop_patience']
    dataloader_num_worker = config['dataloader_num_worker']
    learning_rate = config['learning_rate']
    eps = config['eps']
    eps = float(eps) if not isinstance(eps, float) else eps
    device = config['device']

    #tunable hyperparameters
    hidden_size = config['hidden_size']
    activation = config['activation']

    current_time = time.strftime("%Y%m%d_%H%M%S")
    train_LEM_action(current_time, config, train_path, valid_path, seq_len, features, epochs, batch_size,
                hidden_size,activation,
                num_actions=num_actions,device=device,print_freq=print_freq, patience=early_stop_patience,num_workers=dataloader_num_worker,
                learning_rate=learning_rate,eps=eps)

def main_optuna(config):
    import optuna
    train_path = config['train_path']
    valid_path = config['valid_path']

    # Extract input dimensions from train_df
    if config['test']:
        seq_len = config['seq_len']
        epochs = 2
        features = config['basic_features']+config['other_features']
    else:
        seq_len = config['seq_len']
        epochs = config['num_epoch']
        if config['use_other_features']:
            features = config['basic_features']+config['other_features']
        else:
            features = config['basic_features']
    
    num_actions = config['num_actions']
    print_freq = config['print_freq']
    early_stop_patience = config['early_stop_patience']
    dataloader_num_worker = config['dataloader_num_worker']
    device = config['device']

    def objective(trial):
        # tunable hyperparameters
        try:
            learning_rate=trial.suggest_float('learning_rate', float(config['learning_rate'][0]), float(config['learning_rate'][1])) if len(config['learning_rate'])>1 else config['learning_rate'][0]
            config['eps'] = [float(i) for i in config['eps']]
            eps=trial.suggest_float('eps',  float(config['eps'][1]), float(config['eps'][1])) if len(config['eps'])>1 else config['eps'][0]
            batch_size=trial.suggest_categorical('batch_size', config['batch_size']) 

            hidden_layers = trial.suggest_categorical('hidden_layers', config['hidden_layers'])
            hidden_size = []
            for i in range(hidden_layers):
                hidden_size.append(trial.suggest_categorical(f'hidden_size_{i}', config['hidden_size']))
            activation = trial.suggest_categorical('activation', config['activation'])
        except:
            print("Error in the yaml file")
            
        # Call your train function with these parameters
        valid_loss = train_LEM_action(current_time, config, train_path, valid_path, seq_len, features, epochs, batch_size,
                hidden_size, activation, optuna=True,
                num_actions=num_actions,device=device,print_freq=print_freq, patience=early_stop_patience,num_workers=dataloader_num_worker,
                learning_rate=learning_rate,eps=eps)
        
        return valid_loss


    n_trials = config['optuna_n_trials'] if not config['test'] else 2

    # Create an Optuna study object
    study = optuna.create_study(sampler=optuna.samplers.TPESampler(),direction='minimize')
    
    # Optimize the objective function
    current_time = time.strftime("%Y%m%d_%H%M%S")
    study.optimize(objective, n_trials=n_trials)
    
    # Print the best trial results
    print("Best trial:")
    trial = study.best_trial
    best_trial_number = trial.number+1
    best_params = trial.params
    print(f"  Trial Number: {best_trial_number}")
    print(f"  Best Valid Loss: {trial.value:.4f}")
    print("  Params: ")
    for key, value in trial.params.items():
        print(f"    {key}: {value}")
    
    # Add the trial number to the parameters dictionary
    best_params_with_trial = {
        'best_trial_number': best_trial_number,
        'best_valid_loss': trial.value,
        **best_params
    }

    # Save the best hyperparameters with the trial number to a JSON file
    with open(config['save_path']+f'out/optuna/{current_time}/optuna_best_hyperparameters.json', 'w') as f:
        json.dump(best_params_with_trial, f, indent=4)

if __name__ == "__main__":
    import pandas as pd
    import argparse
    import yaml

    prase = argparse.ArgumentParser()
    prase.add_argument('--config_path', '-c', type=str, default=os.getcwd()+'/event/sports/soccer/models/model_yaml_test/train_LEM_action.yaml')
    args = prase.parse_args()
    
    config_path = args.config_path

    with open(config_path, 'r') as file:
        config = yaml.safe_load(file)

    if config['optuna']:
        main_optuna(config)
    else:
        main(config)

    '''
    all_features = ['match_id','poss_id','team','home_team','action','success','goal','home_score',
                'away_score','goal_diff','Period','Minute','Second','seconds','delta_T',
                'start_x','start_y','deltaX','deltaY','distance','dist2goal','angle2goal']
    basic_features = ['action', 'delta_T', 'start_x','start_y'] 
    derivable_features = ['team','home_team','action','success','seconds','delta_T','start_x','start_y','deltaX','deltaY','distance','dist2goal','angle2goal']
    other_features = ['team','home_team','success','seconds','deltaX','deltaY','distance','dist2goal','angle2goal']
    '''
    



import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.nn.init as init
import math
import numpy as np
import pandas as pd
import json
import os
import time

from ..dataloaders.UEID_loader import UEID_loader
from ..trainers.MAJ_train import MAJ_train
from ..utils.UEID_preprocessing import UEID_preprocessing

import pdb

class MAJ(nn.Module):
    # the model is a simple majority class classifier
    def __init__(self, bin_list_action, bin_list_delta_T, bin_list_start_x, bin_list_start_y):
        super(MAJ, self).__init__()
        # Initialize layers based on input dimensions
        self.action_dist = bin_list_action
        self.delta_T_dist = bin_list_delta_T
        self.start_x_dist = bin_list_start_x
        self.start_y_dist = bin_list_start_y

    
    def forward(self, X):
        batch_size = X.size(0) 
        # return the majority class batch size times
        action = torch.tensor([self.action_dist]*batch_size)
        delta_T = torch.tensor([self.delta_T_dist]*batch_size)
        start_x = torch.tensor([self.start_x_dist]*batch_size)
        start_y = torch.tensor([self.start_y_dist]*batch_size)
        #combine the tensors into one tensor
        out = torch.cat((action, delta_T, start_x, start_y), dim=1)

        return out

         

def get_prob_dist(df,feature, bin=9):
    """
    Get the probability distribution of the features
    """
    bin_list = [0]*bin
    if feature == 'action':
        for i in range(bin):
            bin_list[i] = len(df[df['action']==i])/len(df)
    else:
        df[feature] = (df[feature]*100).astype(int)
        for i in range(bin):
            bin_list[i] = len(df[df[feature]==i])/len(df)

    return bin_list
  
def train_MAJ(current_time, config, train_df, valid_df, seq_len, features, epochs=1, batch_size=256, 
            num_actions=9, device=None, print_freq=10, patience=5, num_workers=4):

    # check if the train_df and valid_df are pandas dataframe
    if not isinstance(train_df, pd.DataFrame):
        #check if the train_df is a path
        if isinstance(train_df, str):
            train_df = pd.read_csv(train_df) if not config['test'] else pd.read_csv(train_df).head(1000)
        else:
            raise ValueError("train_df must be a pandas dataframe or a path to a csv file")
    if not isinstance(valid_df, pd.DataFrame):
        #check if the valid_df is a path
        if isinstance(valid_df, str):
            valid_df = pd.read_csv(valid_df) if not config['test'] else pd.read_csv(valid_df).head(1000)
        else:
            raise ValueError("valid_df must be a pandas dataframe or a path to a csv file")
        
    # Preprocess the data
    train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    valid_df, _, _ = UEID_preprocessing(valid_df, min_dict, max_dict)

    # Get the probability distribution of the features
    bin_list_action = get_prob_dist(train_df, 'action', num_actions)
    bin_list_delta_T = get_prob_dist(train_df, 'delta_T', config['delta_T_bin'])
    bin_list_start_x = get_prob_dist(train_df, 'start_x', config['start_x_bin'])
    bin_list_start_y = get_prob_dist(train_df, 'start_y', config['start_y_bin'])

    #if not device is given, check if cuda is available
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device

    # Initialize model layers based on data
    model = MAJ(bin_list_action, bin_list_delta_T, bin_list_start_x, bin_list_start_y).to(device)
    
    # Create data loaders
    train_loader = UEID_loader(train_df, seq_len, features)
    valid_loader = UEID_loader(valid_df, seq_len, features)

    # Convert data loaders to torch DataLoader
    train_loader = DataLoader(train_loader, batch_size=batch_size, shuffle=True, drop_last=True, num_workers=num_workers)
    valid_loader = DataLoader(valid_loader, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=num_workers)
    

    # Training logic here
    model_state_dict, best_epoch, train_losses, valid_losses, train_loss_components, valid_loss_components = MAJ_train(model, train_loader, valid_loader, min_dict, max_dict, device, epochs, print_freq, patience,config=config)

    #create one csv file for the training and validation losses
    columns = ['train_loss', 'BCEL_action', 'BCEL_deltaT', 'BCEL_start_x', 'BCEL_start_y', 'ACC_action', 'F1_action', 'AE_deltaT', 'AE_start_x', 'AE_start_y',
                'valid_loss', 'BCEL_action_v', 'BCEL_deltaT_v', 'BCEL_start_x_v', 'BCEL_start_y_v', 'ACC_action_v', 'F1_action_v', 'AE_deltaT_v', 'AE_start_x_v', 'AE_start_y_v']
    data = [train_losses, np.array(train_loss_components)[:,0], np.array(train_loss_components)[:,1], np.array(train_loss_components)[:,2], np.array(train_loss_components)[:,3], np.array(train_loss_components)[:,4], np.array(train_loss_components)[:,5], np.array(train_loss_components)[:,6], np.array(train_loss_components)[:,7], np.array(train_loss_components)[:,8],
            valid_losses, np.array(valid_loss_components)[:,0], np.array(valid_loss_components)[:,1], np.array(valid_loss_components)[:,2], np.array(valid_loss_components)[:,3], np.array(valid_loss_components)[:,4], np.array(valid_loss_components)[:,5], np.array(valid_loss_components)[:,6], np.array(valid_loss_components)[:,7], np.array(valid_loss_components)[:,8]]
    data = np.array(data).T
    loss_df = pd.DataFrame(data, columns=columns)
    #round the loss_df to 4 decimal places
    loss_df = loss_df.round(4)

    #save the model and the loss_df
    method = "train"

    i=1
    save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    while os.path.exists(save_path):
        i+=1
        save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    os.makedirs(save_path)

    model_save_path = save_path + f"_model_{best_epoch}.pth"
    loss_save_path = save_path + "_loss.csv"
    hyperparameters_save_path = save_path + "hyperparameters.json"

    #save the min_dict and max_dict
    min_max_dict_path = config['save_path']+f"/out/{method}/{current_time}/min_max_dict.json"
    with open(min_max_dict_path, 'w') as f:
        json.dump({'min_dict':min_dict, 'max_dict':max_dict}, f, indent=4)
    
    #save all the hyperparameters values
    hyperparameters = {'current_time': current_time,
                       'train_path': config['train_path'], 'valid_path': config['valid_path'], 'save_path': config['save_path'],
                        'test': config['test'], 'num_epochs': epochs, 'print_freq': print_freq, 
                        'dataloader_num_worker': config['dataloader_num_worker'], 'batch_size': batch_size, 
                        'device': config['device'], 'basic_features': config['basic_features'], 'features': features,
                        'other_features': [],
                        'num_actions': num_actions, 
                        'seq_len': config['seq_len'],
                        'delta_T_bin': config['delta_T_bin'], 'start_x_bin': config['start_x_bin'], 'start_y_bin': config['start_y_bin'],
                        'bin_list_action': bin_list_action, 'bin_list_delta_T': bin_list_delta_T, 'bin_list_start_x': bin_list_start_x, 'bin_list_start_y': bin_list_start_y,
                        'model':model.__class__.__name__,
                        'best_epoch':best_epoch}
                
    # torch.save(model_state_dict, model_save_path)
    loss_df.to_csv(loss_save_path, index=False)
    with open(hyperparameters_save_path, 'w') as f:
        json.dump(hyperparameters, f, indent=4)
        
    print("Model and loss saved at", save_path)

def main(config):
    train_path = config['train_path']
    valid_path = config['valid_path']

    # Extract input dimensions from train_df
    if config['test']:
        seq_len = 3
        epochs = 1
        features = config['basic_features']
    else:
        seq_len = config['seq_len']
        epochs = config['num_epoch']
        features = config['basic_features']
    
    batch_size = config['batch_size']
    num_actions = config['num_actions']
    print_freq = config['print_freq']
    device = config['device']
    num_workers = config['dataloader_num_worker']

    current_time = time.strftime("%Y%m%d_%H%M%S")
    train_MAJ(current_time, config, train_path, valid_path, seq_len, features, epochs, batch_size, 
            num_actions, device, print_freq, num_workers)
    
if __name__ == "__main__":
    import argparse
    import yaml

    prase = argparse.ArgumentParser()
    prase.add_argument('--config_path', '-c', type=str, default=os.getcwd()+'/event/sports/soccer/models/model_yaml_test/train_MAJ.yaml')
    args = prase.parse_args()
    
    config_path = args.config_path

    with open(config_path, 'r') as file:
        config = yaml.safe_load(file)

    main(config)


import pandas as pd
import numpy as np
import matplotsoccer
import matplotlib.pyplot as plt
import scipy
import json
import os
import pdb  


def plot_heat_map(inference_data_path, min_max_dict_path, save_path, row_num):
    inference_data = pd.read_csv(inference_data_path)

    with open(min_max_dict_path, 'r') as f:
        min_max_dict = json.load(f)
    min_dict= min_max_dict['min_dict']
    max_dict= min_max_dict['max_dict']

    required_col_y = [f'start_y_{i}_prob' for i in range(101)]
    required_col_x = [f'start_x_{i}_prob' for i in range(101)]

    y_bins = np.linspace(0, 100, 101)/100
    x_bins = np.linspace(0, 100, 101)/100

    #scale the bins with the min_max_dict
    y_bins = y_bins*(max_dict["start_y"]-min_dict["start_y"]) + min_dict["start_y"]
    x_bins = x_bins*(max_dict["start_x"]-min_dict["start_x"]) + min_dict["start_x"]

    #check if the required columns are present in the data
    if not set(required_col_y+required_col_x).issubset(inference_data.columns):
        raise ValueError("Missing columns in the data")

    row_sums_x = inference_data[required_col_x].sum(axis=1)
    row_sums_y = inference_data[required_col_y].sum(axis=1)
    if not np.allclose(row_sums_x, 1):
        exp_values = np.exp(inference_data[required_col_x])
        inference_data[required_col_x] = exp_values.div(exp_values.sum(axis=1), axis=0)
    if not np.allclose(row_sums_y, 1):
        exp_values = np.exp(inference_data[required_col_y])
        inference_data[required_col_y] = exp_values.div(exp_values.sum(axis=1), axis=0)

    #fill NaN values with 0
    inference_data.fillna(0, inplace=True)

    y_prob = inference_data[required_col_y].values[row_num]
    x_prob = inference_data[required_col_x].values[row_num]

    #based on the probabilities, sample the x and y coordinates proportionally
    x_data = np.random.choice(x_bins, 100000, p=x_prob)
    y_data = np.random.choice(y_bins, 100000, p=y_prob)

    # matplotsoccer.field("green",figsize=8, show=False)
    # plt.scatter(x_data, y_data, s=1, c="red")
    # plt.axis("on")
    # plt.savefig(save_path+"sc_plot.png")
    # plt.close()

    hm = matplotsoccer.count( pd.Series(x_data), pd.Series(y_data),n=101,m=101) # Construct a 25x25 heatmap from x,y-coordinates
    hm = scipy.ndimage.gaussian_filter(hm,1) # blur the heatmap
    fig, ax = plt.subplots()
    matplotsoccer.heatmap(hm)
    plt.savefig(save_path+"heatmap.png")
    plt.close(fig)

if __name__ == "__main__":
    min_max_dict_path = os.getcwd() + "/test/model/LEM/out/optuna/20241009_010549/min_max_dict.json"
    inference_data_path = os.getcwd() + "/test/inference/LEM/inference.csv"
    save_path = os.getcwd()+"/test/application/"
    row_num = 100
    plot_heat_map(inference_data_path, min_max_dict_path, save_path, row_num)


import pandas as pd

data_path='/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/data.csv'

df = pd.read_csv(data_path)

#get the unique values of the action and team columns
unique_actions = df["action"].unique()
unique_teams = df["team"].unique()

#map the unique values to integers in a dictionary and print the dictionary
action2index = {action: index for index, action in enumerate(unique_actions)}
team2index = {team: index for index, team in enumerate(unique_teams)}
print("Action to index mapping: ", action2index)
print("Team to index mapping: ", team2index)

'''
Action to index mapping:  {'short_pass': 0, 'carry': 1, 'high_pass': 2, '_': 3, 'cross': 4, 'long_pass': 5, 'shot': 6, 'dribble': 7, 'period_over': 8, 'game_over': 9}
Team to index mapping:  {'Granada': 0, 'Atlético Madrid': 1, 'Celta Vigo': 2, 'Osasuna': 3, 'Valencia': 4, 'Sevilla': 5, 'Mallorca': 6, 'Las Palmas': 7, 'Cádiz': 8, 'Deportivo Alavés': 9, 'Real Sociedad': 10, 'Girona': 11, 'Villarreal': 12, 'Real Betis': 13, 'Barcelona': 14, 'Getafe': 15, 'Rayo Vallecano': 16, 'Almería': 17, 'Athletic Club': 18, 'Real Madrid': 19}
'''


import yaml
import json
import os

from ..models.FMS import main as FMS_main
from ..models.FMS import main_optuna  as FMS_main_optuna 
from ..models.LEM_action import main as LEM_action_main
from ..models.LEM_action import main_optuna as LEM_action_main_optuna
from ..models.LEM import main as LEM_main
from ..models.LEM import main_optuna as LEM_main_optuna
from ..models.MAJ import main as MAJ_main
from ..models.NMSTPP import main as NMSTPP_main
from ..models.NMSTPP import main_optuna as NMSTPP_main_optuna
from ..models.Seq2Event import main as Seq2Event_main
from ..models.Seq2Event import main_optuna as Seq2Event_main_optuna

from ..inference import FMS_inference as FMS_inference
from ..inference import LEM_inference as LEM_inference
from ..inference import UEID_inference as UEID_inference

import pdb

class event_model_soccer:
    def __init__(self, model_name, config=None):
        self.model_name = model_name   
        self.config = config
        with open(self.config, 'r') as file:
            config = yaml.safe_load(file)
        self.config = config
        if self.model_name != 'MAJ':
            self.optuna = config['optuna']
        else:
            self.optuna = False


    def train(self):
        if self.optuna:
            if self.model_name == 'FMS':
                FMS_main_optuna(self.config)
            elif self.model_name == 'LEM_action':
                LEM_action_main_optuna(self.config)
            elif self.model_name == 'LEM':
                LEM_main_optuna(self.config)
            elif self.model_name == 'NMSTPP':
                NMSTPP_main_optuna(self.config)
            elif self.model_name == 'Seq2Event':
                Seq2Event_main_optuna(self.config)
            else:
                raise ValueError(f'Unknown model name: {self.model_name}')
        else:
            if self.model_name == 'FMS':
                FMS_main(self.config)
            elif self.model_name == 'LEM_action':
                LEM_action_main(self.config)
            elif self.model_name == 'LEM':
                LEM_main(self.config)
            elif self.model_name == 'MAJ':
                MAJ_main(self.config)
            elif self.model_name == 'NMSTPP':
                NMSTPP_main(self.config)
            elif self.model_name == 'Seq2Event':
                Seq2Event_main(self.config)
            else:
                raise ValueError(f'Unknown model name: {self.model_name}')

    def inference(self, model_path, model_config, train_path=None, valid_path=None, save_path=None, simulation=False, random_selection=True, max_iter=20, min_max_dict_path=None):
        #read the json file
        with open(model_config, 'r') as file:
            config_json = json.load(file)

        #check if the paths are provided
        if train_path is None:
            if 'train_path' in config_json:
                train_path = config_json['train_path']
            elif min_max_dict_path is not None:
                valid_path = self.config['valid_path']
            else:
                raise ValueError('train_path or min_max_dict_path must be provided')
        if valid_path is None:
            valid_path = config_json['valid_path']
        if save_path is None:
            save_path = config_json['save_path']+"/inference/"
        os.makedirs(save_path, exist_ok=True)
        
        #inference function
        if simulation==False:
            if self.model_name in ['FMS', 'MAJ']:
                inferenced_data, loss_df = FMS_inference.FMS_inference(train_path, valid_path, self.model_name, model_path, model_config,min_max_dict_path=min_max_dict_path)
            elif self.model_name=='LEM':
                inferenced_data, loss_df = LEM_inference.LEM_inference(train_path, valid_path, self.model_name, model_path, model_config, min_max_dict_path=min_max_dict_path)
            elif self.model_name in ['NMSTPP', 'Seq2Event']:
                inferenced_data, loss_df = UEID_inference.UEID_inference(train_path, valid_path, self.model_name, model_path, model_config, min_max_dict_path=min_max_dict_path)
            else:
                raise ValueError(f'Unknown model name: {self.model_name}')
            #save the inferenced data
            inferenced_data.to_csv(save_path+"inference.csv",index=False)
            loss_df.to_csv(save_path+"loss.csv",index=False)
            return inferenced_data, loss_df
        else:
            if self.model_name in ['FMS', 'MAJ']:
                df = FMS_inference.FMS_simulation_possession(train_path, valid_path, self.model_name, 
                                                             model_path, model_config, random_selection=random_selection, max_iter=max_iter, min_max_dict_path=min_max_dict_path)
                timestep_eval_df,es_hota_df = FMS_inference.simulation_evaluation(df, valid_path)
            elif self.model_name == 'LEM':
                df = LEM_inference.LEM_simulation_possession(train_path, valid_path, self.model_name, 
                                                             model_path, model_config, random_selection=random_selection, max_iter=max_iter, min_max_dict_path=min_max_dict_path)
                timestep_eval_df,es_hota_df = LEM_inference.simulation_evaluation(df, valid_path)
            elif self.model_name in ['NMSTPP', 'Seq2Event']:
                df = UEID_inference.UEID_simulation_possession(train_path, valid_path, self.model_name, 
                                                             model_path, model_config, random_selection=random_selection, max_iter=max_iter, min_max_dict_path=min_max_dict_path)
                timestep_eval_df,es_hota_df = UEID_inference.simulation_evaluation(df, valid_path)
            else:
                raise ValueError(f'Unknown model name: {self.model_name}')
            #save the inferenced data
            df.to_csv(save_path+"simulation.csv",index=False)
            timestep_eval_df.to_csv(save_path+"timestep_eval.csv",index=False)
            es_hota_df.to_csv(save_path+"ES_HOTA.csv",index=False)
            return df, timestep_eval_df, es_hota_df


if __name__ == '__main__':
    # #Test FMS
    # model = event_model_soccer('FMS', os.getcwd()+'/event/sports/soccer/models/train_FMS.yaml')
    # model.train()
    # #Example only, run the inference function after training
    # model_path = os.getcwd()+'/test/model/FMS/out/train/20240922-181450/run_1/_model_1.pth'
    # model_config = os.getcwd()+'/test/model/FMS/out/train/20240922-181450/run_1/hyperparameters.json'
    # model.inference(model_path, model_config) #simple inference
    # model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20) #simulation with evaluation
    # model = event_model_soccer('FMS', os.getcwd()+'/event/sports/soccer/models/train_FMS_optuna.yaml')
    # model.train()
    # print('FMS Done')

    # #Test LEM_action
    # model = event_model_soccer('LEM_action', os.getcwd()+'/event/sports/soccer/models/train_LEM_action.yaml')
    # model.train()
    # model = event_model_soccer('LEM_action', os.getcwd()+'/event/sports/soccer/models/train_LEM_action_optuna.yaml')
    # model.train()
    # print('LEM_action Done')

    #Test LEM
    # model = event_model_soccer('LEM', os.getcwd()+'/event/sports/soccer/models/train_LEM.yaml')
    # model.train()
    # #Example only, run the inference function after training
    # model_path = os.getcwd()+'/test/model/LEM/out/train/20240922-230404/run_1/_model_1.pth'
    # model_config = os.getcwd()+'/test/model/LEM/out/train/20240922-230404/run_1/hyperparameters.json'
    # model.inference(model_path, model_config) #simple inference
    # model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20) #simulation with evaluation
    # model = event_model_soccer('LEM', os.getcwd()+'/event/sports/soccer/models/train_LEM_optuna.yaml')
    # model.train()
    # print('LEM Done')
    
    # #Test MAJ
    # model = event_model_soccer('MAJ', os.getcwd()+'/event/sports/soccer/models/train_MAJ.yaml')
    # model.train()
    # model_path = os.getcwd()+'/test/model/MAJ/out/train/20240923_061427/run_1/_model_1.pth'
    # model_config = os.getcwd()+'/test/model/MAJ/out/train/20240923_061427/run_1/hyperparameters.json'
    # model.inference(model_path, model_config) #simple inference
    # model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20) #simulation with evaluation
    # print('MAJ Done')

    # #Test NMSTPP
    # model = event_model_soccer('NMSTPP', os.getcwd()+'/event/sports/soccer/models/train_NMSTPP.yaml')
    # model.train()
    # model_path = os.getcwd()+'/test/model/NMSTPP/out/train/20240922-134432/run_1/_model_1.pth'
    # model_config = os.getcwd()+'/test/model/NMSTPP/out/train/20240922-134432/run_1/hyperparameters.json'
    # model.inference(model_path, model_config) #simple inference
    # model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20) #simulation with evaluation
    # model = event_model_soccer('NMSTPP', os.getcwd()+'/event/sports/soccer/models/train_NMSTPP_optuna.yaml')
    # model.train()
    # print('NMSTPP Done')

    # #Test Seq2Event
    # model = event_model_soccer('Seq2Event', os.getcwd()+'/event/sports/soccer/models/train_Seq2Event.yaml')
    # model.train()
    # model_path = os.getcwd()+'/test/model/Seq2Event/out/train/20240922-150633/run_1/_model_1.pth'
    # model_config = os.getcwd()+'/test/model/Seq2Event/out/train/20240922-150633/run_1/hyperparameters.json'
    # model.inference(model_path, model_config) #simple inference
    # model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20) #simulation with evaluation
    # model = event_model_soccer('Seq2Event', os.getcwd()+'/event/sports/soccer/models/train_Seq2Event_optuna.yaml')
    # model.train()
    # print('Seq2Event Done')

    #Test inference FMS
    # model = event_model_soccer('FMS', os.getcwd()+'/event/sports/soccer/models/model_yaml_test/train_FMS_inference.yaml')
    # # model.train()
    # #Example only, run the inference function after training
    # model_path = os.getcwd()+'/test/model/FMS/out/train/20240928_114700/run_2/_model_1.pth'
    # model_config = os.getcwd()+'/test/model/FMS/out/train/20240928_114700/run_2/hyperparameters.json'
    # min_max_dict_path = os.getcwd()+'/test/model/FMS/out/train/20240928_114700/min_max_dict.json'
    # model.inference(model_path, model_config, min_max_dict_path=min_max_dict_path) #simple inference
    # model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20, min_max_dict_path=min_max_dict_path) #simulation with evaluation
    # print('FMS Done')

    print('Done')


import torch
import numpy as np
from ..utils.FMS_cost_function import FMS_cost_function
import torchprofile
import warnings
import pdb

def FMS_epoch_function(model, dataloader, min_dict, max_dict, optimizer, device, train=True, print_freq=10,config=None):
    """
    Function to run one epoch of training or evaluation.
    
    Parameters:
    - model: The neural network model.
    - dataloader: DataLoader for the dataset.
    - optimizer: Optimizer for training (if train=True).
    - train: Boolean, True for training, False for evaluation.
    - print_freq: Frequency of printing the loss.

    Returns:
    - Average loss for the epoch.
    """
    if train:
        model.train()
    else:
        model.eval()

    total_loss = 0
    total_loss_components = [0]*9
    total_batches = len(dataloader)
    
    for i, (inputs, gt, _) in enumerate(dataloader):
        
        inputs, gt = inputs.to(device), gt.to(device)
        
        optimizer.zero_grad() if train else None

        # Forward pass
        outputs = model(inputs)
    
        loss, loss_components= FMS_cost_function(gt, outputs,  min_dict=min_dict, max_dict=max_dict, device=device,config=config)

        if train:
            # Backward pass and optimization
            loss.backward()
            optimizer.step()

        total_loss += loss.item()
        #detach the loss components
        loss_components = [i.cpu().detach().numpy() for i in loss_components] if device != "cpu" else [i.detach().numpy() for i in loss_components]
        total_loss_components = [total_loss_components[j] + loss_components[j] for j in range(len(loss_components))]

        if (i + 1) % print_freq == 0 or (i + 1) == total_batches:

            average_loss = total_loss / (i + 1)
            average_loss_components = [total_loss_components[j] / (i + 1) for j in range(len(loss_components))]
            
            print(f"Batch [{i + 1}/{total_batches}] | Loss: {average_loss:.4f} | "
                    f"CEL_action: {average_loss_components[0]:.4f} | CEL_deltaT: {average_loss_components[1]:.4f} | "
                    f"CEL_start_x: {average_loss_components[2]:.4f} | CEL_start_y: {average_loss_components[3]:.4f} | "
                    f"ACC_action: {average_loss_components[4]:.4f} | "
                    f"F1_action: {average_loss_components[5]:.4f} | "
                    f"MAE_deltaT: {average_loss_components[6]:.4f} | MAE_x: {average_loss_components[7]:.4f} | "
                    f"MAE_y: {average_loss_components[8]:.4f}")

    average_loss = total_loss / total_batches

    if train:
        return average_loss, average_loss_components, model.state_dict()
    else:
        return average_loss, average_loss_components

def FMS_train(model, train_loader, valid_loader, min_dict, max_dict, optimizer, device, epochs=10, print_freq=10, patience=5,config=None):
    """
    Function to train the model with both training and validation datasets, with early stopping.

    Parameters:
    - model: The neural network model.
    - train_loader: DataLoader for the training dataset.
    - valid_loader: DataLoader for the validation dataset.
    - mean_dict: Dictionary for the mean of each feature.
    - std_dict: Dictionary for the standard deviation of each feature.
    - optimizer: Optimizer for training.
    - epochs: Number of epochs to train.
    - print_freq: Frequency of printing the loss during training.
    - patience: Number of epochs with no improvement after which training will be stopped.

    Returns:
    - model: The trained model.
    - train_losses: List of training losses for each epoch.
    - valid_losses: List of validation losses for each epoch.
    - train_loss_components: List of training loss components for each epoch.
    - valid_loss_components: List of validation loss components for each epoch.
    """
    torch.cuda.empty_cache(); import gc; gc.collect()
    train_losses = []
    valid_losses = []
    train_loss_components = []
    valid_loss_components = []

    best_valid_loss = float('inf')
    epochs_no_improve = 0
    best_epoch = 0

    for epoch in range(epochs):
        torch.cuda.empty_cache(); import gc; gc.collect()
        print(f"Epoch [{epoch + 1}/{epochs}]")

        # Training phase
        train_loss, average_loss_components, model_state_dict = FMS_epoch_function(model, train_loader, min_dict, max_dict, optimizer, device, train=True, print_freq=print_freq,config=config)
        train_losses.append(train_loss)
        train_loss_components.append(average_loss_components)
        print(f"Training Loss: {train_loss:.4f}")

        # Validation phase
        with torch.no_grad():
            torch.cuda.empty_cache(); import gc; gc.collect()
            valid_loss, average_loss_components = FMS_epoch_function(model, valid_loader, min_dict, max_dict, optimizer=None, device=device, train=False, print_freq=print_freq,config=config)
        valid_losses.append(valid_loss)
        valid_loss_components.append(average_loss_components)
        print(f"Validation Loss: {valid_loss:.4f}\n")

        # Check if validation loss improved
        if valid_loss < best_valid_loss:
            best_valid_loss = valid_loss
            best_model = model_state_dict
            best_epoch = epoch
            epochs_no_improve = 0
        else:
            epochs_no_improve += 1

        # Early stopping
        if epochs_no_improve >= patience:
            print(f"Early stopping triggered after {epoch + 1} epochs.")
            model_state_dict = best_model
            epoch = best_epoch
            break

    #Print the best epoch, the best validation loss and the corresponding training loss
    print(f"Best epoch: {best_epoch + 1} | Best validation loss: {best_valid_loss:.4f} | Corresponding training loss: {train_losses[best_epoch]:.4f}")

    warnings.filterwarnings("ignore", message="No handlers found: .*")
    def get_flops(model, dataloader, device='cpu'):
        # Move model to the specified device
        model.to(device)
        
        # Get a batch of data from the DataLoader and move it to the same device
        inputs, _, _ = next(iter(dataloader))
        inputs = inputs.to(device)
        
        # Ensure the model is in evaluation mode
        model.eval()
        
        # Profile FLOPs based on the provided input data
        flops = torchprofile.profile_macs(model, inputs)
        
        # Multiply by 2 (if needed for your specific calculation)
        flops = flops * 2
        
        return flops
    
    # Get the FLOPs for the model
    flops = get_flops(model, train_loader, device=device)

    #get the best model number of parameters
    best_model = model_state_dict
    num_params = sum(p.numel() for p in best_model.values())

    model_stats = {"flops":flops, "num_params":num_params}
    print(f"FLOPs: {flops} | Number of parameters: {num_params}")
    
    return model_state_dict, epoch, train_losses, valid_losses, train_loss_components, valid_loss_components, model_stats




import torch
import torch.nn as nn
import pdb

def LEM_action_cost_function(gt, pred, device=None,config=None):

    num_actions = config["num_actions"] if config is not None and config !="None" else 9
    gt_action = gt[:,0:num_actions].float()

    #Binary Cross Entropy Loss
    BCEL_action_loss = torch.nn.BCELoss()
    BCEL_action = torch.mean(BCEL_action_loss(pred,gt_action))

    #Accuracy
    ACC_action = torch.mean((torch.argmax(pred,1)==torch.argmax(gt_action,1)).float())

    #F1 Score
    f1_action = calculate_f1_score(torch.argmax(gt_action,1),torch.argmax(pred,1),num_actions)

    return BCEL_action, [BCEL_action, ACC_action,f1_action]

def calculate_f1_score(gt_classes, pred_classes, num_classes):
    f1_scores = []
    
    for i in range(num_classes):
        # True Positives (TP), False Positives (FP), False Negatives (FN)
        tp = torch.sum((pred_classes == i) & (gt_classes == i)).float()
        fp = torch.sum((pred_classes == i) & (gt_classes != i)).float()
        fn = torch.sum((pred_classes != i) & (gt_classes == i)).float()
        
        # Precision and Recall
        precision = tp / (tp + fp + 1e-8)  # Add small constant to avoid division by zero
        recall = tp / (tp + fn + 1e-8)     # Add small constant to avoid division by zero
        
        # F1 Score
        f1 = 2 * (precision * recall) / (precision + recall + 1e-8)  # Add small constant to avoid division by zero
        f1_scores.append(f1)
    
    # Average F1 Score across all classes
    avg_f1_score = torch.mean(torch.stack(f1_scores))
    
    return avg_f1_score

def LEM_cost_function(gt, pred,min_dict=None, max_dict=None, device="None",config=None):
    num_actions = config["num_actions"] if config is not None and config !="None" else 9
    num_deltaT = config["delta_T_bin"] if config is not None and config !="None" else 61
    num_start_x = config["start_x_bin"] if config is not None and config !="None" else 101
    num_start_y = config["start_y_bin"] if config is not None and config !="None" else 101
    gt_action = gt[:,0:num_actions].float()
    gt_deltaT = gt[:,num_actions:num_actions+num_deltaT].float()
    gt_start_x = gt[:,num_actions+num_deltaT:num_actions+num_deltaT+num_start_x].float()
    gt_start_y = gt[:,num_actions+num_deltaT+num_start_x:].float()
    pred_action = pred[:,0:num_actions]
    pred_deltaT = pred[:,num_actions:num_actions+num_deltaT]
    pred_start_x = pred[:,num_actions+num_deltaT:num_actions+num_deltaT+num_start_x]
    pred_start_y = pred[:,num_actions+num_deltaT+num_start_x:]
    #BCEL
    BCEL_continuous_loss = torch.nn.BCELoss()   
    BCEL_continuous = torch.mean(BCEL_continuous_loss(pred[:,num_actions:],gt[:,num_actions:]))

    #AE for deltaT, start_x, start_y
    if max_dict is not None and min_dict is not None:
        deltaT_pred = torch.argmax(pred_deltaT,1)/100
        deltaT_pred = torch.exp( deltaT_pred * ( torch.tensor(max_dict["delta_T"], device=device) - torch.tensor(min_dict["delta_T"], device=device)) + torch.tensor(min_dict["delta_T"], device=device)) - torch.tensor(1e-6, device=device) 
        deltaT = torch.argmax(gt_deltaT,1)/100
        deltaT = torch.exp( deltaT * ( torch.tensor(max_dict["delta_T"], device=device) - torch.tensor(min_dict["delta_T"], device=device)) + torch.tensor(min_dict["delta_T"], device=device)) - torch.tensor(1e-6, device=device)
        start_x_pred = torch.argmax(pred_start_x,1)/100
        start_x_pred = start_x_pred * ( torch.tensor(max_dict["start_x"], device=device) - torch.tensor(min_dict["start_x"], device=device)) + torch.tensor(min_dict["start_x"], device=device)
        start_x = torch.argmax(gt_start_x,1)/100
        start_x = start_x * ( torch.tensor(max_dict["start_x"], device=device) - torch.tensor(min_dict["start_x"], device=device)) + torch.tensor(min_dict["start_x"], device=device)
        start_y_pred = torch.argmax(pred_start_y,1)/100
        start_y_pred = start_y_pred * ( torch.tensor(max_dict["start_y"], device=device) - torch.tensor(min_dict["start_y"], device=device)) + torch.tensor(min_dict["start_y"], device=device)
        start_y = torch.argmax(gt_start_y,1)/100
        start_y = start_y * ( torch.tensor(max_dict["start_y"], device=device) - torch.tensor(min_dict["start_y"], device=device)) + torch.tensor(min_dict["start_y"], device=device)
        
        #ignore in backpropagation
        AE_deltaT = torch.mean(torch.abs(deltaT_pred-deltaT)).detach()
        AE_start_x = torch.mean(torch.abs(start_x_pred-start_x)).detach()
        AE_start_y = torch.mean(torch.abs(start_y_pred-start_y)).detach()
    else:
        AE_deltaT = torch.tensor(-1, device=device)
        AE_start_x = torch.tensor(-1, device=device)
        AE_start_y = torch.tensor(-1, device=device)

    return BCEL_continuous, [BCEL_continuous,AE_deltaT,AE_start_x,AE_start_y]


import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.nn import functional as F
import math
import numpy as np
import pandas as pd
import json
import os
import time

from ..dataloaders.UEID_loader import UEID_loader
from ..trainers.UEID_train import train
from ..utils.UEID_preprocessing import UEID_preprocessing

import pdb

class Seq2Event(nn.Module):
    def __init__(self, action_embedding_input_len, action_embedding_out_len, scale_grad_by_freq, 
                 continuous_embedding_input_len, continuous_embedding_output_len,
                 multihead_attention, hidden_dim,
                 transformer_num_layers,transformer_finaldenselayer_dim,
                 deltaT_output_len=1, location_output_len=2, action_output_len=9):
        super(Seq2Event, self).__init__()
        
        # Initialize layers based on input dimensions
        self.action_embedding = nn.Embedding(action_embedding_input_len, action_embedding_out_len, scale_grad_by_freq=scale_grad_by_freq)
        self.continuous_embedding = nn.Linear(continuous_embedding_input_len, continuous_embedding_output_len, bias=True)
        self.d_model=action_embedding_out_len+continuous_embedding_output_len
        self.positional_encoding = PositionalEncoding(d_model=self.d_model)

        # Transformer encoder layer
        self.encoder_layer = nn.TransformerEncoderLayer(d_model=self.d_model, nhead=multihead_attention, batch_first=True, dim_feedforward=hidden_dim)
        self.seqnet = nn.TransformerEncoder(self.encoder_layer, num_layers=transformer_num_layers)
        
        # Prediction NN layers
        self.lin1 = nn.Linear(self.d_model,transformer_finaldenselayer_dim)
        self.lin2 = nn.Linear(transformer_finaldenselayer_dim, deltaT_output_len+location_output_len+action_output_len,bias=True)
    
    def forward(self, X):
        #input features 
        X_action=X[:,:,0]
        X_continuous=X[:,:,1:]

        #embedding
        X_action = self.action_embedding(X_action.int())
        X_continuous= self.continuous_embedding(X_continuous.float())
        X_concatenate = torch.cat((X_action,X_continuous),2).float()

        #positional encoding
        X_positional = X_concatenate+ self.positional_encoding(X_concatenate)
        X_positional=X_positional.float()
        
        #encoder layer
        X_encode= self.seqnet(X_positional)

        #prediction layers
        out = self.lin1(X_encode[:,-1,:])
        out = F.relu(out)
        out = self.lin2(out)

        return out
    
class PositionalEncoding(nn.Module):

    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        
        # Ensure div_term has the correct size
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        
        # Fill in the positional encoding
        pe[:, 0::2] = torch.sin(position * div_term)
        if d_model % 2 == 1:  # Handle odd d_model by adding a column of zeros
            pe[:, 1::2] = torch.cos(position * div_term[:d_model//2])
        else:
            pe[:, 1::2] = torch.cos(position * div_term)

        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)

def train_Seq2Event(current_time, config, train_df, valid_df, seq_len, features, epochs, batch_size,
            action_embedding_out_len, scale_grad_by_freq, 
            continuous_embedding_output_len,
            multihead_attention, hidden_dim, transformer_num_layers,transformer_finaldenselayer_dim,
            deltaT_output_len=1, location_output_len=2, action_output_len=9,
            num_actions=9, device=None, print_freq=10, patience=5, num_workers=4, optuna=False):

    # check if the train_df and valid_df are pandas dataframe
    if not isinstance(train_df, pd.DataFrame):
        #check if the train_df is a path
        if isinstance(train_df, str):
            train_df = pd.read_csv(train_df) if not config['test'] else pd.read_csv(train_df).head(1000)
        else:
            raise ValueError("train_df must be a pandas dataframe or a path to a csv file")
    if not isinstance(valid_df, pd.DataFrame):
        #check if the valid_df is a path
        if isinstance(valid_df, str):
            valid_df = pd.read_csv(valid_df) if not config['test'] else pd.read_csv(valid_df).head(1000)
        else:
            raise ValueError("valid_df must be a pandas dataframe or a path to a csv file")
        
    # Preprocess the data
    train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    valid_df, _, _ = UEID_preprocessing(valid_df, min_dict, max_dict)
    
    # Define fixed hyperparameters 
    num_continous_features = len(features) - 1
    action_embedding_input_len=num_actions
    continuous_embedding_input_len=num_continous_features

    deltaT_output_len=1
    location_output_len=2
    action_output_len=num_actions

    #if not device is given, check if cuda is available
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device

    # Initialize model layers based on data
    model = Seq2Event(action_embedding_input_len, action_embedding_out_len, scale_grad_by_freq, 
                continuous_embedding_input_len, continuous_embedding_output_len,
                multihead_attention, hidden_dim,
                transformer_num_layers,transformer_finaldenselayer_dim,
                deltaT_output_len, location_output_len, action_output_len).to(device)

    # Create data loaders
    train_loader = UEID_loader(train_df, seq_len, features)
    valid_loader = UEID_loader(valid_df, seq_len, features)

    # Convert data loaders to torch DataLoader
    train_loader = DataLoader(train_loader, batch_size=batch_size, shuffle=True, drop_last=True, num_workers=num_workers)
    valid_loader = DataLoader(valid_loader, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=num_workers)
    
    # Define optimizer
    optimizer = torch.optim.Adam(model.parameters(),lr=0.01,eps=1e-16)

    # Training logic here
    model_state_dict, best_epoch, train_losses, valid_losses, train_loss_components, valid_loss_components, flops = train(model, train_loader, valid_loader, min_dict, max_dict, optimizer, device, epochs, print_freq, patience,config=config)

    #create one csv file for the training and validation losses
    columns = ['train_loss', 'CEL_action', 'RMSE_deltaT', 'RMSE_location', 'ACC_action', 'F1_action', 'MAE_deltaT', 'MAE_x', 'MAE_y',
                'valid_loss', 'CEL_action_v', 'RMSE_deltaT_v', 'RMSE_location_v', 'ACC_action_v', 'F1_action_v', 'MAE_deltaT_v', 'MAE_x_v', 'MAE_y_v']
    data = [train_losses, np.array(train_loss_components)[:,0], np.array(train_loss_components)[:,1], np.array(train_loss_components)[:,2], np.array(train_loss_components)[:,3], np.array(train_loss_components)[:,4], np.array(train_loss_components)[:,5], np.array(train_loss_components)[:,6], np.array(train_loss_components)[:,7],
            valid_losses, np.array(valid_loss_components)[:,0], np.array(valid_loss_components)[:,1], np.array(valid_loss_components)[:,2], np.array(valid_loss_components)[:,3], np.array(valid_loss_components)[:,4], np.array(valid_loss_components)[:,5], np.array(valid_loss_components)[:,6], np.array(valid_loss_components)[:,7]]
    data = np.array(data).T
    loss_df = pd.DataFrame(data, columns=columns)
    #round the loss_df to 4 decimal places
    loss_df = loss_df.round(4)

    
    #save the model and the loss_df
    if config['optuna']:
        method = "optuna"
    else:
        method = "train"
    i=1
    save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    while os.path.exists(save_path):
        i+=1
        save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    os.makedirs(save_path)

    model_save_path = save_path + f"_model_{best_epoch}.pth"
    loss_save_path = save_path + "_loss.csv"
    hyperparameters_save_path = save_path + "hyperparameters.json"
    flops_save_path = save_path + "_flops.txt"

    #save the min_dict and max_dict
    min_max_dict_path = config['save_path']+f"/out/{method}/{current_time}/min_max_dict.json"
    with open(min_max_dict_path, 'w') as f:
        json.dump({'min_dict':min_dict, 'max_dict':max_dict}, f, indent=4)
    
    #save all the hyperparameters values
    hyperparameters = {'current_time': current_time,
                       'train_path': config['train_path'], 'valid_path': config['valid_path'], 'save_path': config['save_path'],
                        'test': config['test'], 'batch_size': batch_size, 'num_epoch': config['num_epoch'],
                        'print_freq': print_freq, 'early_stop_patience': config['early_stop_patience'],
                        'dataloader_num_worker': config['dataloader_num_worker'], 'device': config['device'], 
                        'basic_features': config['basic_features'], 'other_features': config['other_features'],
                        'use_other_features': config['use_other_features'],
                        'features':features, 'num_actions': num_actions, 'seq_len': seq_len,
                        'optuna': config['optuna'], 'optuna_n_trials': config['optuna_n_trials'],
                        'num_continous_features': num_continous_features, 'action_embedding_input_len': action_embedding_input_len,
                        'action_embedding_out_len': action_embedding_out_len, 'scale_grad_by_freq': scale_grad_by_freq,
                        'continuous_embedding_input_len': continuous_embedding_input_len,
                        'continuous_embedding_output_len': continuous_embedding_output_len,
                        'multihead_attention': multihead_attention, 'hidden_dim': hidden_dim,
                        'transformer_num_layers': transformer_num_layers, 
                        'transformer_finaldenselayer_dim': transformer_finaldenselayer_dim,
                        'deltaT_output_len':deltaT_output_len, 
                        'location_output_len':location_output_len, 
                        'action_output_len':action_output_len,
                        'model':model.__class__.__name__,
                        'best_epoch':best_epoch}
                
    torch.save(model_state_dict, model_save_path)
    loss_df.to_csv(loss_save_path, index=False)
    with open(hyperparameters_save_path, 'w') as f:
        json.dump(hyperparameters, f, indent=4)
    with open(flops_save_path, "w") as f:
        f.write(str(flops))
        

    print("Model and loss saved at", save_path)

    if optuna:
        return valid_losses[best_epoch]

def main(config):
    train_path = config['train_path']
    valid_path = config['valid_path']

    # Extract input dimensions from train_df
    if config['test']:
        seq_len = 10
        epochs = 2
        features = config['basic_features']+config['other_features']
    else:
        seq_len = config['seq_len']
        epochs = config['num_epoch']
        if config['use_other_features']:
            features = config['basic_features']+config['other_features']
        else:
            features = config['basic_features']
    
    batch_size = config['batch_size']
    num_actions = config['num_actions']
    print_freq = config['print_freq']
    early_stop_patience = config['early_stop_patience']
    dataloader_num_worker = config['dataloader_num_worker']
    device = config['device']

    #tunable hyperparameters
    action_embedding_out_len = config['action_embedding_out_len']
    scale_grad_by_freq = config['scale_grad_by_freq']
    continuous_embedding_output_len = config['continuous_embedding_output_len']
    
    multihead_attention = config['multihead_attention'] #fix to 1 given the previous papers
    hidden_dim = config['hidden_dim']
    transformer_num_layers = config['transformer_num_layers']
    transformer_finaldenselayer_dim = config['transformer_finaldenselayer_dim']

    current_time = time.strftime("%Y%m%d_%H%M%S")
    train_Seq2Event(current_time, config, train_path, valid_path, seq_len, features, epochs, batch_size,
                action_embedding_out_len, scale_grad_by_freq, 
                continuous_embedding_output_len,
                multihead_attention, hidden_dim,transformer_num_layers,transformer_finaldenselayer_dim,
                num_actions=num_actions,device=device,print_freq=print_freq, patience=early_stop_patience,num_workers=dataloader_num_worker)

def main_optuna(config):
    import optuna
    train_path = config['train_path']
    valid_path = config['valid_path']

    # Extract input dimensions from train_df
    if config['test']:
        seq_len = 10
        epochs = 2
        features = config['basic_features']+config['other_features']
    else:
        seq_len = config['seq_len']
        epochs = config['num_epoch']
        if config['use_other_features']:
            features = config['basic_features']+config['other_features']
        else:
            features = config['basic_features']
    
    batch_size = config['batch_size']
    num_actions = config['num_actions']
    print_freq = config['print_freq']
    early_stop_patience = config['early_stop_patience']
    dataloader_num_worker = config['dataloader_num_worker']
    device = config['device']

    def objective(trial):
        #tunable hyperparameters
        try:
            action_embedding_out_len=trial.suggest_categorical('action_embedding_out_len', config['action_embedding_out_len'])
            scale_grad_by_freq=trial.suggest_categorical('scale_grad_by_freq', config['scale_grad_by_freq'])
            continuous_embedding_output_len=trial.suggest_categorical('continuous_embedding_output_len', config['continuous_embedding_output_len'])
            
            multihead_attention=1 #fix to 1 given the previous papers
            hidden_dim=trial.suggest_categorical('hidden_dim', config['hidden_dim'])
            transformer_num_layers=trial.suggest_categorical('transformer_num_layers', config['transformer_num_layers'])
            transformer_finaldenselayer_dim=trial.suggest_categorical('transformer_finaldenselayer_dim', config['transformer_finaldenselayer_dim'])
        except:
            print("Error in the yaml file")
            
        # Call your train function with these parameters
        valid_loss = train_Seq2Event(current_time, config, train_path, valid_path, seq_len, features, epochs, batch_size,
                action_embedding_out_len, scale_grad_by_freq, 
                continuous_embedding_output_len,
                multihead_attention, hidden_dim,transformer_num_layers,transformer_finaldenselayer_dim, optuna=True,
                num_actions=num_actions,device=device,print_freq=print_freq, patience=early_stop_patience,num_workers=dataloader_num_worker)
        
        return valid_loss


    n_trials = config['optuna_n_trials'] if not config['test'] else 2

    # Create an Optuna study object
    study = optuna.create_study(direction='minimize')
    
    # Optimize the objective function
    current_time = time.strftime("%Y%m%d_%H%M%S")
    study.optimize(objective, n_trials=n_trials)
    
    # Print the best trial results
    print("Best trial:")
    trial = study.best_trial
    best_trial_number = trial.number+1
    best_params = trial.params
    print(f"  Trial Number: {best_trial_number}")
    print(f"  Best Valid Loss: {trial.value:.4f}")
    print("  Params: ")
    for key, value in trial.params.items():
        print(f"    {key}: {value}")
    
    # Add the trial number to the parameters dictionary
    best_params_with_trial = {
        'best_trial_number': best_trial_number,
        'best_valid_loss': trial.value,
        **best_params
    }

    # Save the best hyperparameters with the trial number to a JSON file
    with open(config['save_path']+f'out/optuna/{current_time}/optuna_best_hyperparameters.json', 'w') as f:
        json.dump(best_params_with_trial, f, indent=4)

if __name__ == "__main__":
    import pandas as pd
    import argparse
    import yaml

    prase = argparse.ArgumentParser()
    prase.add_argument('--config_path', '-c', type=str, default=os.getcwd()+'/event/sports/soccer/models/model_yaml_test/train_Seq2Event.yaml')
    args = prase.parse_args()
    
    config_path = args.config_path

    with open(config_path, 'r') as file:
        config = yaml.safe_load(file)

    if config['optuna']:
        main_optuna(config)
    else:
        main(config)

    '''
    all_features = ['match_id','poss_id','team','home_team','action','success','goal','home_score',
                'away_score','goal_diff','Period','Minute','Second','seconds','delta_T',
                'start_x','start_y','deltaX','deltaY','distance','dist2goal','angle2goal']
    basic_features = ['action', 'delta_T', 'start_x','start_y'] 
    derivable_features = ['team','home_team','action','success','seconds','delta_T','start_x','start_y','deltaX','deltaY','distance','dist2goal','angle2goal']
    other_features = ['team','home_team','success','seconds','deltaX','deltaY','distance','dist2goal','angle2goal']
    '''
    



import os
import json
from tqdm import tqdm
import pandas as pd
from statsbombpy import sb

from preprocessing import Event_data

#down_load_statsbomb_data function
def download_statsbomb_data(creds, save_dir,competition_id=11, season_id=281):

    os.makedirs(save_dir, exist_ok=True)

    def convert_df_in_dict(d):
        for key, value in d.items():
            if isinstance(value, pd.DataFrame):
                d[key] = value.to_dict(orient='records')
            elif isinstance(value, dict):
                convert_df_in_dict(value)
        return d

    # Get Statsbomb matches data
    matches = sb.matches(competition_id=competition_id, season_id=season_id, creds=creds)
    matches["competition_id"] = competition_id
    matches["season_id"] = season_id
    #moev the competition_id and season_id to the first column
    cols = matches.columns.tolist()
    cols = cols[-2:] + cols[:-2]
    matches = matches[cols]
    #save the matches to csv
    matches.to_csv(os.path.join(save_dir, "matches.csv"), index=False)

    # Get Statsbomb lineups and events
    os.makedirs(os.path.join(save_dir, "lineups"), exist_ok=True)
    os.makedirs(os.path.join(save_dir, "events"), exist_ok=True)
    for match_id in tqdm(matches["match_id"].unique()):
        lineups = sb.lineups(match_id=match_id, creds=creds)
        events = sb.events(match_id=match_id, include_360_metrics=True, creds=creds)
        events.to_csv(os.path.join(save_dir, "events", f"{match_id}.csv"), index=False)
        #save the lineups as json and with row changes
        lineups = convert_df_in_dict(lineups)
        with open(os.path.join(save_dir, "lineups", f"{match_id}.json"), "w") as f:
            json.dump(lineups, f, indent=4)

if __name__ == "__main__":
    #Statsbomb API
    # creds = {"user": "input your Statsbomb api user name here", "passwd": "input your Statsbomb api password here"}
    #Statsbomb event data saving dir
    # save_dir = "/home/c_yeung/workspace6/python/openstarlab/EventModeling/test/dataset/"
    #path to the skillcorner tracking data
    tracking_path="/skillcorner/tracking"
    #path to the skillcorner match data
    match_path="/skillcorner/match"

    statsbomb_skillcorner_event_path="/data_pool_1/laliga_23/statsbomb/events"
    statsbomb_skillcorner_tracking_path="/data_pool_1/laliga_23/skillcorner/tracking"
    statsbomb_skillcorner_match_path="/data_pool_1/laliga_23/skillcorner/match"

    # download_statsbomb_data(creds, save_dir)

    #output file
    match_id_df_path="/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/id_matching.csv"
    output_dir="/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/data.csv"

    #Match the statsbomb and skillcorner (multiple files)
    statsbomb_skillcorner_df=Event_data(data_provider='statsbomb_skillcorner',
                                        statsbomb_event_dir=statsbomb_skillcorner_event_path,
                                        skillcorner_tracking_dir=statsbomb_skillcorner_tracking_path,
                                        skillcorner_match_dir=statsbomb_skillcorner_match_path,
                                        match_id_df=match_id_df_path, 
                                        preprocess_method="UIED"
                                        ).preprocessing()
    statsbomb_skillcorner_df.to_csv(output_dir,index=False)
    print("---------------done-----------------")

    


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import pdb

def cal_HPUS(data, shot_num=[6,8], cross_num=[4], num_actions=9):
   # Check if the input is a DataFrame or CSV file path
   if not isinstance(data, pd.DataFrame):
      data = pd.read_csv(data)

   # Map actions to integers and handle goals (action = 8 for goal)
   action_dict = {'short_pass': 0, 'carry': 1, 'high_pass': 2, '_': 3, 'cross': 4, 
                  'long_pass': 5, 'shot': 6, 'dribble': 7}
   data["action"] = data["action"].map(action_dict)
   data.loc[(data["action"] == 6) & (data["goal"] == 1), "action"] = 8

   # Remove rows with 'period_over' (action = 3)
   data = data[data["action"] != 3]

   # Fill NaN values with 0 for all columns
   data.fillna(0, inplace=True)

   # Define zone score based on `start_x_pred_unscaled`
   conditions = [
      (data["start_x_pred_unscaled"] > 83 * 1.05),
      (data["start_x_pred_unscaled"] < 50 * 1.05),
      ((50 * 1.05 <= data["start_x_pred_unscaled"]) & (data["start_x_pred_unscaled"] <= 83 * 1.05))
   ]
   zone_scores = [10, 0, 5]
   data["zone_score"] = np.select(conditions, zone_scores, default=0)

   # Normalize action probabilities if needed
   action_columns = [f'action_{i}_prob' for i in range(num_actions)]
   row_sums = data[action_columns].sum(axis=1)
   if not np.allclose(row_sums, 1):
      exp_values = np.exp(data[action_columns])
      data[action_columns] = exp_values.div(exp_values.sum(axis=1), axis=0)

   # Calculate action score based on weighted action probabilities
   weights = [5, 5, 5, 0, 10, 5, 10, 5, 10]  # Weights corresponding to each action
   data['action_score'] = sum(weights[i] * data[f'action_{i}_prob'] for i in range(num_actions))

   # Compute HPUS_t and HPUS
   data['HPUS_t'] = data['HPUS_t'] = np.clip(data['delta_T_pred_unscaled'], 1, None)
   data['HPUS'] = np.sqrt(data['action_score'] * data['zone_score']) / data['HPUS_t']

   # Compute attack_flag
   relevant_actions = set(shot_num + cross_num)
   data['attack_flag'] = data['action'].isin(relevant_actions).astype(int)

   # Group by match_id and poss_id to compute HPUS values
   HPUS_value = data.groupby(['match_id', 'poss_id']).apply(
    lambda group: pd.Series({
        'HPUS': np.sum(group['HPUS'] * np.exp(-0.3 * (len(group) - np.arange(len(group))))),
        'team': group['team'].iloc[0],
        'seconds': group['seconds'].iloc[-1],
        'attack_flag': int(group['attack_flag'].sum() > 0)
    })
   ).reset_index()

   return HPUS_value

def plot_HPUS(hpus_data, save_path, match_id=None, plus=False, time_period=5):
    if match_id is None:
        match_id = hpus_data['match_id'].unique()[0]

    # Filter the data for the given match_id
    match_data = hpus_data[hpus_data['match_id'] == match_id]

    if plus:
        # Filter the data for attack_flag = 1
        match_data = match_data[match_data['attack_flag'] == 1]
    
    # Create the plot
    plt.figure(figsize=(10, 6))

    # Iterate over each unique team in the match
    for team in match_data['team'].unique():
        # Filter data for the specific team
        team_data = match_data[match_data['team'] == team].copy()

        # Create 5-minute bins for the `seconds` column (intervals: 0-5, 5-10, ..., up to 110)
        bins = list(range(0, 110 * 60 + 1, time_period * 60))
        team_data['time_bin'] = pd.cut(team_data['seconds'], bins=bins, labels=bins[:-1]).to_list()

        # Aggregate the HPUS values per 5-minute interval
        y = team_data.groupby('time_bin')['HPUS'].sum()

        # Check for NaN values and replace them with 0 if necessary
        y = y.fillna(0)

        # Prepare x values as time in minutes
        x = [bin_val / 60 for bin_val in y.index.astype(int)]

        # Plot for each team
        plt.plot(x, y, marker='o', linestyle='-', label=f'Team {team}')

    # Add titles and labels
    if not plus:
        plt.title(f'HPUS Plot for Match {match_id}', fontsize=18) 
        plt.ylabel('HPUS', fontsize=18) 
    else:
        plt.title(f'HPUS+ Plot for Match {match_id}', fontsize=18) 
        plt.ylabel('HPUS+', fontsize=18) 

    plt.xlabel('Match Time (minutes)', fontsize=18)

    # Increase font size for tick labels
    plt.tick_params(axis='both', labelsize=16)

    # Add grid and legend
    plt.grid(True)
    plt.legend(fontsize=16)

    # Save the plot
    if not plus:
        plt.savefig(f"{save_path}/HPUS.png")
    else:
        plt.savefig(f"{save_path}/HPUS_plus.png")
    
    plt.close()  # Close the figure to free memory



if __name__ == "__main__":
   import os
   inference_data = os.getcwd()+"/test/inference/nmstpp/inference.csv"
   save_path = os.getcwd()+"/test/application/"
   hpus=cal_HPUS(inference_data)
   hpus.to_csv(save_path+"HPUS.csv", index=False)
   plot_HPUS(hpus,save_path)
   plot_HPUS(hpus,save_path,plus=True)
   print("HPUS and HPUS+ plots saved successfully.")
   pdb.set_trace()

    


import torch
import numpy as np
from ..utils.MAJ_cost_function import MAJ_cost_function
import torchprofile
import warnings
import pdb

def MAJ_epoch_function(model, dataloader, min_dict, max_dict, device, train=True, print_freq=10,config=None):
    model.eval()

    total_loss = 0
    total_loss_components = [0]*9
    total_batches = len(dataloader)
    
    for i, (inputs, gt, _) in enumerate(dataloader):
        inputs, gt = inputs.to(device), gt.to(device)
        num_action=config["num_actions"] if config is not None or config !="None" else 9
        
        gt_action_one_hot = torch.nn.functional.one_hot(gt[:,0].long(), num_classes=num_action).float()
        gt_deltaT_one_hot = one_hot_encode_from_continuous(gt[:,1], num_bins=config["delta_T_bin"]).float()
        gt_start_x_one_hot = one_hot_encode_from_continuous(gt[:,2], num_bins=config["start_x_bin"]).float()
        gt_start_y_one_hot = one_hot_encode_from_continuous(gt[:,3], num_bins=config["start_y_bin"]).float()
        gt_binary = torch.cat((gt_action_one_hot, gt_deltaT_one_hot, gt_start_x_one_hot, gt_start_y_one_hot), dim=1)

        # Forward pass
        outputs = model(inputs)
        loss, loss_components= MAJ_cost_function(gt_binary, outputs, min_dict, max_dict, device=device,config=config)
        
        total_loss += loss.item()
        #detach the loss components
        loss_components = [i.cpu().detach().numpy() for i in loss_components] if device != "cpu" else [i.detach().numpy() for i in loss_components]
        total_loss_components = [total_loss_components[j] + loss_components[j] for j in range(len(loss_components))]

        if (i + 1) % print_freq == 0 or (i + 1) == total_batches:

            average_loss = total_loss / (i + 1)
            average_loss_components = [total_loss_components[j] / (i + 1) for j in range(len(loss_components))]
            
            print(f"Batch [{i + 1}/{total_batches}] | Loss: {average_loss:.4f} | "
                    f"BCEL_action: {average_loss_components[0]:.4f} | BCEL_deltaT: {average_loss_components[1]:.4f} | "
                    f"BCEL_start_x: {average_loss_components[2]:.4f} | BCEL_start_y: {average_loss_components[3]:.4f} | " 
                    f"ACC_action: {average_loss_components[4]:.4f} | F1_action: {average_loss_components[5]:.4f} | "
                    f"AE_deltaT: {average_loss_components[6]:.4f} | AE_start_x: {average_loss_components[7]:.4f} | AE_start_y: {average_loss_components[8]:.4f}"
                    )

    average_loss = total_loss / total_batches

    if train:
        return average_loss, average_loss_components, model.state_dict()
    else:
        return average_loss, average_loss_components

def one_hot_encode_from_continuous(gt, num_bins):
    # Clamp the continuous values to stay between 0 and 1
    gt = torch.clamp(gt, 0, 1)
    
    # Map the continuous values to bin indices
    bin_idx = (gt * (num_bins - 1)).long()

    # One-hot encode the bin indices
    one_hot_encoded = torch.nn.functional.one_hot(bin_idx, num_classes=num_bins).float()

    return one_hot_encoded

def MAJ_train(model, train_loader, valid_loader, min_dict, max_dict,  device, epochs=10, print_freq=10, patience=5,config=None):
    torch.cuda.empty_cache(); import gc; gc.collect()
    train_losses = []
    valid_losses = []
    train_loss_components = []
    valid_loss_components = []

    best_valid_loss = float('inf')
    epochs_no_improve = 0
    best_epoch = 0

    for epoch in range(epochs):
        torch.cuda.empty_cache(); import gc; gc.collect()
        print(f"Epoch [{epoch + 1}/{epochs}]")

        with torch.no_grad():
            # Training phase
            train_loss, average_loss_components, model_state_dict = MAJ_epoch_function(model, train_loader, min_dict, max_dict, device, train=True, print_freq=print_freq,config=config)
            train_losses.append(train_loss)
            train_loss_components.append(average_loss_components)
            print(f"Training Loss: {train_loss:.4f}")

            # Validation phase
            torch.cuda.empty_cache(); import gc; gc.collect()
            valid_loss, average_loss_components = MAJ_epoch_function(model, valid_loader, min_dict, max_dict, device=device, train=False, print_freq=print_freq,config=config)
        valid_losses.append(valid_loss)
        valid_loss_components.append(average_loss_components)
        print(f"Validation Loss: {valid_loss:.4f}\n")

        # Check if validation loss improved
        if valid_loss < best_valid_loss:
            best_valid_loss = valid_loss
            best_model = model_state_dict
            best_epoch = epoch
            epochs_no_improve = 0
        else:
            epochs_no_improve += 1

        # Early stopping
        if epochs_no_improve >= patience:
            print(f"Early stopping triggered after {epoch + 1} epochs.")
            model_state_dict = best_model
            epoch = best_epoch
            break

    #Print the best epoch, the best validation loss and the corresponding training loss
    print(f"Best epoch: {best_epoch + 1} | Best validation loss: {best_valid_loss:.4f} | Corresponding training loss: {train_losses[best_epoch]:.4f}")


    return model_state_dict, epoch, train_losses, valid_losses, train_loss_components, valid_loss_components, 




from event import Event_Model

import os

# To train NMSTPP model
model = Event_Model('NMSTPP', 'path/to/train_NMSTPP.yaml')
# #or to use optuna for hyperparameter optimization
# model = Event_Model('NMSTPP', 'path/to/train_NMSTPP_optuna.yaml')
model.train()

# Inference
# Example only, run the inference function after training
model_path = 'path/to/_model_1.pth'
model_config = 'path/to/hyperparameters.json'
# Simple inference
model.inference(model_path, model_config) 
# Simulation with evaluation
model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20) 

